from pydantic import validate_call
"""
Some docs here
"""
from mezcla import debug


def not_called_function() ->None:
    """This function is not called"""
    print('This function is not called')
    print(validate_call(debug.get_level)())


def move_debug_level() ->None:
    """Move debug level"""
    print('Default debug level: ', debug.DEFAULT)
    print('Current debug level: ', validate_call(debug.get_level)())
    validate_call(debug.set_level)(validate_call(debug.get_level)() + 1)
    validate_call(debug.set_level)(validate_call(debug.get_level)() - 1)


def main() ->None:
    """Main function"""
    print('Hello, ...')
    validate_call(move_debug_level)()
    print('... World!')


if __name__ == '__main__':
    validate_call(main)()



#! /usr/bin/env python3
#
# Run embedded JavaScript from HTML page through code checking tools, such as
# jslint and jshint. This includes common definitions such as for DOM object
# (e.g., window) and JQuery ($ selector function). It also optionally runs
# in strict mode to help check other potential errors (e.g., undefined vars).
#
#

"""Run JavaScript embedded in <script> tags through lint-style code checkers"""

# Standard modules
from collections import defaultdict
import re
import tempfile

# Local modules
# TODO: def tomas_import(name): ... components = eval(name).split(); ... import nameN-1.nameN as nameN
from mezcla import debug
from mezcla import glue_helpers as gh
from mezcla.main import Main
from mezcla.my_regex import my_re
from mezcla import system

CODE_CHECKERS = "code-checkers"
STRIP_INDENT = "strip-indent"
SKIP_SAFE_MODE = "skip-safe-mode"
SKIP_COMMON_DEFINES = "skip-common-defines"
JAVASCRIPT_HEADER = "javascript-header"
#
# TODO: put script-based temporary filename basename support into main
DEFAULT_PREFIX = (system.remove_extension(__file__) or "check-js")
TEMP_PREFIX = system.getenv_text("TEMP_PREFIX", DEFAULT_PREFIX,
                                 "Temporary file prefix (see tempfile.mkstemp)")
TEMP_SUFFIX = system.getenv_text("TEMP_SUFFIX", "-",
                                 "Temporary file suffix (see tempfile.mkstemp)")
# NOTE: see tempfile.mkstemp for NamedTemporaryFile keyword args
DEFAULT_TEMP_BASE = tempfile.NamedTemporaryFile(prefix=TEMP_PREFIX, suffix=TEMP_SUFFIX).name
TEMP_BASE = system.getenv_text("TEMP_BASE", DEFAULT_TEMP_BASE,
                               "Basename with directory for temporary files")
#
MAX_ERRORS = system.getenv_int("MAX_ERRORS", 10000,
                               "Maxmium number of erros to report")
#
# note: see DEFAULT_JAVASCRIPT_HEADER below for other options
JSLINT = "jslint"
JSLINT_OPTIONS = system.getenv_text("JSLINT_OPTIONS", f"--maxerr {MAX_ERRORS} --white",
                                    "Options for jslint")
JSHINT = "jshint"
JSHINT_OPTIONS = system.getenv_text("JSHINT_OPTIONS", "--show-non-errors",
                                    "Options for jshint")
DEFAULT_CODE_CHECKERS = system.getenv_text("CODE_CHECKERS",
                                           f"{JSLINT},  {JSHINT}",
                                           "JavaScript code checking commands")
SAFEMODE_HEADER = """
'use strict';            // Added for sanity checking (e.g., undefined variables)
"""

# TODO: use separate headers for jslint and jshint
DEFAULT_JAVASCRIPT_HEADER = f"""
// Start of added header (JavaScript and jQuery definitions)
function $(selector, context) {{ selector = context; }}
var document;
var window;
var jQuery;

// Stuff for jslint:
//    global $, jQuery, alert
//    jslint browser: true
//    jslint devel          // Allow console.log() and friends.
//    jslint long           // Allow long lines.
//    jslint white          // Allow messy whitespace.

// Stuff for jshint:
//    jshint maxerr: {MAX_ERRORS}

// End of added header
"""
# TODO3: add pointer to definition (e.g., https://www.jslint.com)

class Script(Main):
    """Input processing class"""
    # TODO: -or-: """Adhoc script class (e.g., no I/O loop, just run calls)"""
    code_checkers = DEFAULT_CODE_CHECKERS
    strip_indent = False
    skip_safe_mode = False
    skip_common_defines = False
    javascript_header = DEFAULT_JAVASCRIPT_HEADER

    def __init__(self, *args, **kwargs):
        debug.trace_fmtd(5, "Script.__init__({a}): keywords={kw}; self={s}",
                         a=",".join(args), kw=kwargs, s=self)
        self.script_code = ""
        self.in_script = False
        self.code_indent = None
        super().__init__(*args, **kwargs)

    def setup(self):
        """Check results of command line processing"""
        debug.trace_fmtd(5, "Script.setup(): self={s}", s=self)
        self.code_checkers = self.get_parsed_option(CODE_CHECKERS, self.code_checkers)
        self.strip_indent = self.get_parsed_option(STRIP_INDENT, self.strip_indent)
        self.skip_safe_mode = self.get_parsed_option(SKIP_SAFE_MODE, self.skip_safe_mode)
        self.skip_common_defines = self.get_parsed_option(SKIP_COMMON_DEFINES, self.skip_common_defines)
        self.javascript_header = self.get_parsed_option(JAVASCRIPT_HEADER, self.javascript_header)
        debug.trace_object(5, self, label="Script instance")

    def process_line(self, line):
        """Processes current line from input"""
        # Notes: Issues warning when src attribute given along with bracketted code
        # Also, issues warning when bracketted code on same line as <script> tag.
        debug.trace_fmtd(6, "Script.process_line({l})", l=line)
        entire_line = line
        script_tag_count = 0

        # Check for start of code section, ignoring external script via src attribute
        while my_re.search(r"^([^<>]*)<\/?script[^<>]*>(.*)", line):
            remainder = my_re.group(2)
            if my_re.search(r"^\s*<script[^<>]*>(.*)$", line):
                debug.assertion(not self.in_script)
                self.in_script = True
                line = my_re.group(1)
                script_tag_count += 1
            elif my_re.search(r"^(.*)</script[^<>]*>(.*)", line):
                debug.assertion(self.in_script)
                code = my_re.group(1).strip()
                line = my_re.group(2)
                self.in_script = False
                if code:
                    self.script_code += code + "\n"
            else:
                system.print_stderr("Warning: ignoring unexpected script tag formatting at line {n}: {t}", n=self.line_num, t=entire_line)
                line = remainder
        debug.assertion(script_tag_count <= 1)
        if script_tag_count:
            debug.trace_fmt(4, "{n} <script> taggings: in_script={ins}", n=script_tag_count, ins=self.in_script)

        # Accumulate text if within script tags
        if self.in_script:
            # Make sure indentation is defined if being stripped
            # Note: All code is stripped of indentation, so that aligns with javascript header.
            # This is just determined by first non-blank line, so that if-blocks, etc. not stripped of all spacing.
            if (self.strip_indent and line.strip()):
                # Make sure indentation defined
                if (self.code_indent is None):
                    self.code_indent = ""
                    if my_re.search(r"^(\s+)(.*)", line):
                        self.code_indent = my_re.group(1)
                        line = my_re.group(2)
                        debug.assertion(len(self.code_indent) > 0)
                    debug.trace_fmt(4, "Indent to strip: '{ind}' (len={l})", ind=self.code_indent, l=len(self.code_indent))
                # Strip indentation from start of line
                line = line.replace(self.code_indent, "", 1)
            #
            self.script_code += line.rstrip() + "\n"

    def wrap_up(self):
        """Run the accumulated script through code checkers"""
        if (not self.script_code.strip()):
            system.print_stderr("Error: No code found within <script> tags")
        javascript_file = (TEMP_BASE + ".js")

        # Add in header for strict mode (optional) and a few JavaScript defines
        code_header = ""
        if not self.skip_safe_mode:
            code_header += SAFEMODE_HEADER
        if not self.skip_common_defines:
            code_header += self.javascript_header
        system.write_file(javascript_file, code_header + self.script_code)
        output = None
        default_options_hash = defaultdict(str)
        default_options_hash.update({JSLINT: JSLINT_OPTIONS,
                                     JSHINT: JSHINT_OPTIONS})
        for checker in re.split(", *", self.code_checkers):
            if output is not None:
                print("-" * 80)
            options_var = f"{checker}_OPTIONS".upper()
            checker_options = system.getenv_text(options_var,
                                                 default_options_hash[checker])
            output = gh.run("{ch} {opt} {scr}",
                            ch=checker, opt=checker_options, scr=javascript_file)
            print("Output from {ch}:".format(ch=checker))
            print(output)
            print("")

#-------------------------------------------------------------------------------
    
if __name__ == '__main__':
    debug.trace_current_context(level=debug.QUITE_DETAILED)
    app = Script(
        description=__doc__,
        # Note: skip_input controls the line-by-line processing, which is inefficient but simple to
        # understand; in contrast, manual_input controls iterator-based input (the opposite of both).
        skip_input=False,
        manual_input=False,
        # TODO: skip_input=True,
        # TODO: manual_input=True,
        boolean_options=[STRIP_INDENT, SKIP_SAFE_MODE, SKIP_COMMON_DEFINES],
        text_options=[
            (CODE_CHECKERS, "Comma-separated list of code checking invocations (e.g., '{dfc}')".format(dfc=DEFAULT_CODE_CHECKERS)),
            (JAVASCRIPT_HEADER, "JavaScript header with common definitions (e.g., document, window, jQuery)")])
    app.run()



#! /usr/bin/env python3
#
# Test(s) for ../corpus.py

"""Tests for tfidf corpus submodule"""

# Standard packages
import math

# Installed modules
import pytest

# Local modules
from mezcla.unittest_wrapper import TestWrapper
from mezcla import debug
from mezcla import system
from mezcla.tfidf.preprocess import Preprocessor

# Note: Two references are used for the module to be tested:
#    THE_MODULE:                        global module object
import mezcla.tfidf.corpus as THE_MODULE

# ------------------------------------------------------------------------


class TestPreprocess(TestWrapper):
    """Class for testcase definition"""
    
    script_module = TestWrapper.get_testing_module_name(__file__, THE_MODULE)
    
    THE_MODULE.NGRAM_EPSILON = 0
    corp = THE_MODULE.Corpus(
        preprocessor=Preprocessor(gramsize=1, language='english')
    )
    corp_2 = THE_MODULE.Corpus(
        preprocessor=Preprocessor(gramsize=1, language='english')
    )
    
    text_1 = "The quick brown fox jumps over the lazy dog. The fox is cunning and swift."
    text_2 = "A lazy cat sleeps in the warm sunlight. The cat ignores the playful dog nearby."
    text_3 = "Dogs and foxes are both members of the canine family. Some dogs are as quick as foxes."
    
    corp["doc_1"] = text_1
    corp["doc_2"] = text_2
    corp["doc_3"] = text_3
    
    @classmethod
    def setUpClass(cls, filename=None, module=None):
        """One-time initialization (i.e., for entire class)"""
        debug.trace(6, f"TestPreprocess.setUpClass(); cls={cls}")
        # note: should do parent processing first
        super().setUpClass(filename, module)
        debug.trace_object(5, cls, label=f"{cls.__class__.__name__} instance")

    @pytest.mark.xfail  # TODO: remove xfail
    def test_01_df_freq(self):
        """test document frequency for ngram"""
        assert self.corp.df_freq('dog') == 3
        assert self.corp.df_freq('fox') == 2
            
    @pytest.mark.xfail  # TODO: remove xfail
    def test_02_df_norm(self):
        """Tests the document freq for an ngram normalized
        by the max doc frequency"""
        assert self.corp.df_norm('dog') == 1
        assert self.corp.df_norm('fox') == 2/3
        
    @pytest.mark.xfail  # TODO: remove xfail
    def test_03_idf_basic(self):
        """Test IDF with basic normalization"""
        assert self.corp.idf_basic('dog') == 0
        assert self.corp.idf_basic('fox') == math.log(3/2)
        
    @pytest.mark.xfail  # TODO: remove xfail
    def test_04_idf_freq(self):
        """Test IDF doc freq"""
        assert self.corp.idf_freq('dog') ==  1 / 3
        assert self.corp.idf_freq('fox') ==  1 / 2

    @pytest.mark.xfail  # TODO: remove xfail
    def test_05_idf_smooth(self):
        """Test IDF with simple smoothing"""
        assert self.corp.idf_smooth('dog') == math.log(2)
        assert self.corp.idf_smooth('fox') == math.log(1 + 3/2)
        
    @pytest.mark.xfail  # TODO: remove xfail
    def test_06_idf_max(self):
        """Test IDF with max TF and add-1 smoothing"""
        assert self.corp.idf_max('dog') == math.log(1 + (2/3))
        assert self.corp.idf_max('fox') == math.log(2)
    
    @pytest.mark.xfail  # TODO: remove xfail
    def test_07_idf_probabilistic(self):
        """Test IDF via probabilistic interpretation"""
        ## OLD:
        ## assert self.corp.idf_probabilistic('dog') == math.log(1)
        ## assert self.corp.idf_probabilistic('fox') == math.log(3/2)
        # Note:
        # IDF_prob = log((N - DF)/DF + epsilon) for N docs and DF doc. freq.
        # dog => log((3 - 3) / 3 + epsilon); fox => log((3 - 2) / 2 + epsilon)
        assert system.round3(self.corp.idf_probabilistic('dog')) == -13.816
        assert system.round3(self.corp.idf_probabilistic('fox')) == -0.693
    
    @pytest.mark.xfail  # TODO: remove xfail
    def test_08_tf_idf(self):
        """Test term frequency/inverse document frequency"""
        assert self.corp.tf_idf('dog', 'doc_3', tf_weight='basic').score == 0
        assert self.corp.tf_idf('fox', 'doc_3', tf_weight='basic').score == 0.25 * math.log(3/2)
        
    @pytest.mark.xfail  # TODO: remove xfail
    def test_09_get_keywords(self):
        """Test ngram keywords with scores"""
        keywords = self.corp.get_keywords('doc_1')
        doc = self.corp['doc_1']
        for keyword in keywords:
            assert keyword.ngram in doc
# ------------------------------------------------------------------------

if __name__ == "__main__":

    debug.trace_current_context()
    ## TODO2: here and elsewhere: invoke_tests(__file__)
    pytest.main([__file__])



#! /usr/bin/env python3
#
# note: Not used in any test.
#

"""
Illustration script to test validation of most used functions from:
- html_utils.py
- debug.py
- system.py

List of most used functions in Mezcla package (✓ == tested on this script, x == not tested):

✓ trace(...) -> called 32 times
x get_exception(...) -> called 21 times
✓ val(...) -> called 17 times
✓ get_url_param(...) -> called 16 times
x trace_fmt(...) -> called 15 times
x code(...) -> called 12 times
x debugging(...) -> called 12 times
x assertion(...) -> called 12 times
✓ get_url_parameter(...) -> called 12 times
x getenv(...) -> called 12 times
x get_url_parameter_value(...) -> called 11 times
x call(...) -> called 9 times
x trace_fmtd(...) -> called 9 times
x format_value(...) -> called 9 times
✓ escape_html_text(...) -> called 9 times
x trace_expr(...) -> called 8 times
x download_web_document(...) -> called 8 times
✓ escape_html_value(...) -> called 8 times
x trace_object(...) -> called 7 times
✓ get_level(...) -> called 7 times
x print_stderr(...) -> called 7 times
x getenv_int(...) -> called 7 times
x _to_utf8(...) -> called 6 times
x get_param_dict(...) -> called 6 times
x getenv_value(...) -> called 6 times
x profile_function(...) -> called 5 times
x timestamp(...) -> called 5 times
x docstring_parameter(...) -> called 5 times
x main(...) -> called 5 times
x get_browser(...) -> called 5 times
✓ open_file(...) -> called 5 times
✓ file_exists(...) -> called 5 times
x to_utf8(...) -> called 5 times
x to_float(...) -> called 5 times
x _getenv_bool(...) -> called 4 times
x _print_exception_info(...) -> called 4 times
x open_debug_file(...) -> called 4 times
x trace_values(...) -> called 4 times
x init_BeautifulSoup(...) -> called 4 times
✓ unescape_html_text(...) -> called 4 times
✓ round_num(...) -> called 4 times
x init(...) -> called 4 times
x register_env_option(...) -> called 4 times
x getenv_number(...) -> called 4 times
x to_int(...) -> called 4 times
x print_error(...) -> called 4 times
x print_full_stack(...) -> called 4 times
✓ write_file(...) -> called 4 times
x to_str(...) -> called 4 times
x to_string(...) -> called 4 times
"""

# Standard packages
import tempfile

# Installed packages
from mezcla.main import Main
from mezcla import system
from mezcla import debug
from mezcla import html_utils

# Command-line labels and
# enviroment variables constants
WRONG = 'wrong'
GOOD = 'good'

class ExampleValidationScript(Main):
    """Argument processing class"""

    # Class-level member variables for arguments
    # (avoids need for class constructor)
    wrong = False
    good = True

    def setup(self) -> None:
        """Process arguments"""
        self.wrong = self.get_parsed_option(WRONG, self.wrong)
        self.good = self.get_parsed_option(GOOD, self.good)

    def run_main_step(self) -> None:
        """Process main script"""
        if self.good:
            print("With GOOD parameter types:")
            print(f"\t{'✓' if self.is_debug_working() else 'x'} mezcla.debug")
            print(f"\t{'✓' if self.is_html_utils_working() else 'x'} mezcla.html_utils")
            print(f"\t{'✓' if self.is_system_working() else 'x'} mezcla.system")
        if self.wrong:
            print("With WRONG parameter types:")
            print(f"\t{'✓' if self.is_debug_failing() else 'x'} mezcla.debug")
            print(f"\t{'✓' if self.is_html_utils_failing() else 'x'} mezcla.html_utils")
            print(f"\t{'✓' if self.is_system_failing() else 'x'} mezcla.system")

    def is_debug_working(self) -> bool:
        """Check if debug functions work with good types"""
        debug.trace(7, 'Works as expected!')
        debug.trace(debug.TraceLevel.QUITE_VERBOSE, 'Works as expected!')
        assert debug.val(debug.get_level(), 'Hello, world!') == 'Hello, world!'
        return True

    def is_html_utils_working(self) -> bool:
        """Check if html_utils functions work with good types"""
        html_utils.get_url_param(
            "some-key",
            "some-value",
            { "some-key": "foobar" },
        )
        html_utils.escape_html_text("<2/")
        html_utils.escape_html_text("Joe's hat")
        return True

    def is_system_working(self) -> bool:
        """Check if system functions work with good types"""
        temp_file = tempfile.NamedTemporaryFile().name
        system.write_file(temp_file, 'Hello, world!')
        assert system.file_exists(temp_file)
        file = system.open_file(temp_file, 'w')
        assert file is not None
        file.close()
        assert system.round_num(1.6789, 2) == 1.68
        return True

    def fails_successfully(self, func, *args, **kwargs) -> bool:
        """Check if function fails as expected"""
        try:
            func(*args, **kwargs)
            return False
        except TypeError: ## TODO: change to specific exception
            return True
        except Exception:
            return False

    def is_debug_failing(self) -> bool:
        """Check if debug functions fail with wrong types"""
        to_check = [
            self.fails_successfully(debug.trace, None, 'foo'),
            self.fails_successfully(debug.trace, "7", 'foo'),
            self.fails_successfully(debug.trace, debug.TraceLevel.QUITE_VERBOSE, {'a': 1, 'b': 2}),
            self.fails_successfully(debug.val, None, 'Hello, world!'),
        ]
        return all(to_check)

    def is_html_utils_failing(self) -> bool:
        """Check if html_utils functions fail with wrong types"""
        to_check = [
            self.fails_successfully(html_utils.get_url_param, "some-key", "some-value", True),
            self.fails_successfully(html_utils.escape_html_text, True),
            self.fails_successfully(html_utils.escape_html_text, 7),
        ]
        return all(to_check)

    def is_system_failing(self) -> bool:
        """Check if system functions fail with wrong types"""
        to_check = [
            self.fails_successfully(system.write_file, None, 'Hello, world!'),
            self.fails_successfully(system.file_exists, None),
            self.fails_successfully(system.write_file, 7, 'Hello, world!'),
            self.fails_successfully(system.file_exists, 7),
            self.fails_successfully(system.round_num, None, 2),
            self.fails_successfully(system.round_num, 1.6789, '2'),
        ]
        return all(to_check)


if __name__ == '__main__':

    app = ExampleValidationScript(
        description = __doc__,
        boolean_options = [
            (WRONG, 'Run methods with wrong types'),
            (GOOD, 'Run methods with good types'),
        ],
        manual_input = True,
    )

    app.run()



#! /usr/bin/env python3
#
# via https://machinelearningmastery.com/how-to-calculate-precision-recall-f1-and-more-for-deep-learning-models/
#
# Note:
# - Sloppy programming by the blogger.
#

"""Illustrates ML metrics using synthetic circles dataset"""

# demonstration of calculating metrics for a neural network model using sklearn
from sklearn.linear_model import LogisticRegression
from sklearn.datasets import make_circles
from sklearn.metrics import accuracy_score
from sklearn.metrics import precision_score
from sklearn.metrics import recall_score
from sklearn.metrics import f1_score
from sklearn.metrics import cohen_kappa_score
from sklearn.metrics import roc_auc_score
from sklearn.metrics import confusion_matrix
from keras.models import Sequential
from keras.layers import Dense

# Local packages
from mezcla import debug
from mezcla import system

# Environment options
USE_KERAS = system.getenv_bool("USE_KERAS", False,
                               "Use Keras deep learning for shites and giggles")
RANDOM_SEED = system.getenv_value("RANDOM_SEED", 15485863,
                                  "Seed to use for random state")

def get_data():
    """generate and prepare the dataset"""
    # generate dataset
    X, y = make_circles(n_samples=1000, noise=0.1, random_state=RANDOM_SEED)
    debug.trace_expr(5, X, y)
    # split into train and test
    n_test = 500
    trainX, testX = X[:n_test, :], X[n_test:, :]
    trainy, testy = y[:n_test], y[n_test:]
    return trainX, trainy, testX, testy


def get_model(trainX, trainy):
    """define and fit the model"""
    # define model
    if USE_KERAS:
        model = Sequential()
        model.add(Dense(100, input_dim=2, activation='relu'))
        model.add(Dense(1, activation='sigmoid'))
        # compile model
        model.compile(loss='binary_crossentropy', optimizer='adam', metrics=['accuracy'])
        # fit model
        model.fit(trainX, trainy, epochs=300, verbose=0)
    else:
        model = LogisticRegression()
        model.fit(trainX, trainy)
    return model


def main():
    """Entry point"""

    # generate data
    trainX, trainy, testX, testy = get_data()
    # fit model
    model = get_model(trainX, trainy)

    # predict probabilities for test set
    if USE_KERAS:
        yhat_probs = model.predict(testX, verbose=0)
    else:
        yhat_probs = model.predict_proba(testX)
    # predict crisp classes for test set
    if USE_KERAS:
        # pylint: disable=no-member
        yhat_classes = model.predict_classes(testX, verbose=0)
    else:
        yhat_classes = model.predict(testX)
    # reduce to 1d array
    yhat_probs = yhat_probs[:, 0]
    ## BAD: yhat_classes = yhat_classes[:, 0]
    if (len(yhat_classes.shape) > 1):
        yhat_classes = yhat_classes[:, 0]
    
    # accuracy: (tp + tn) / (p + n)
    accuracy = accuracy_score(testy, yhat_classes)
    print('Accuracy: %f' % accuracy)
    # precision: tp / (tp + fp)
    precision = precision_score(testy, yhat_classes)
    print('Precision: %f' % precision)
    # recall: tp / (tp + fn)
    recall = recall_score(testy, yhat_classes)
    print('Recall: %f' % recall)
    # f1: 2 tp / (2 tp + fp + fn)
    f1 = f1_score(testy, yhat_classes)
    print('F1 score: %f' % f1)
    
    # kappa
    kappa = cohen_kappa_score(testy, yhat_classes)
    print('Cohens kappa: %f' % kappa)
    # ROC AUC
    auc = roc_auc_score(testy, yhat_probs)
    print('ROC AUC: %f' % auc)
    # confusion matrix
    matrix = confusion_matrix(testy, yhat_classes)
    print(matrix)

#-------------------------------------------------------------------------------

if __name__ == "__main__":
    main()



# Common types used in multiple modules
#
# These types are the same as those used in builtins.pyi,
# but we need to copy them here because they are protected
# and cannot be imported directly.
#
# We need to separate the types used in several functions
# in this module to avoid circular imports.

# TODO: reduce redundancy
"""
Common types used in multiple modules

These types are the same as those used in builtins.pyi,
but we need to copy them here because they are protected
and cannot be imported directly.

We need to separate the types used in several functions
in this module to avoid circular imports.
"""

from os import PathLike
from types import TracebackType
from typing import Union, Tuple

# Types
StrOrBytesPath = Union[str, bytes, PathLike]  # stable
FileDescriptorOrPath = Union[int, StrOrBytesPath]
ExcInfo = Tuple[BaseException, TracebackType]
OptExcInfo = Union[ExcInfo, Tuple[None, None, None]]



#! /usr/bin/env python3
#
# Test(s) for ../test_matrix_multiply_benchmarking.py
#
# Notes:
# - Sample output tested
#   $ SKIP_CUDA=1 ../matrix_multiply_benchmarking.py --numba --numpy
#   numpy	0.304ms
#   numba	11817.419ms
# - This can be run as follows (e.g., from root of repo):
#   $ pytest ./mezcla/examples/tests/test_test_matrix_multiply_benchmarking.py
#

## TODO1: [Warning] Make sure this template adhered to as much as possible. For,
## example, only delete todo comments not regular code, unless suggested in tip).
## In particular, it is critical that script_module gets initialized properly.

"""TODO: Tests for test_matrix_multiply_benchmarking module"""

# Standard modules
## TODO: from collections import defaultdict

# Installed modules
import pytest

# Local modules
from mezcla.unittest_wrapper import TestWrapper
## TODO: from mezcla.unittest_wrapper import trap_exception
from mezcla import debug
## TODO: from mezcla import glue_helpers as gh
from mezcla.my_regex import my_re
from mezcla import system

# Note: Two references are used for the module to be tested:
#    THE_MODULE:                        global module object
#    TestIt.script_module:              path to file
try:
    import mezcla.examples.matrix_multiply_benchmarking as THE_MODULE
except:
    THE_MODULE = None
    system.print_exception_info("matrix_multiply_benchmarking import")
#
# Note: sanity test for customization (TODO: remove if desired)
if not my_re.search(__file__, r"\btemplate.py$"):
    debug.assertion("mezcla.template" not in str(THE_MODULE))

## TODO:
## # Environment options
## # Note: These are just intended for internal options, not for end users.
## # It also allows for enabling options in one place.
## #
## FUBAR = system.getenv_bool(
##     "FUBAR", False,
##     description="Fouled Up Beyond All Recognition processing")

#------------------------------------------------------------------------

class TestIt(TestWrapper):
    """Class for command-line based testcase definition"""
    script_module = TestWrapper.get_testing_module_name(__file__, THE_MODULE)

    @pytest.mark.xfail                   # TODO: remove xfail
    ## DEBUG: @trap_exception            # TODO: remove when debugged
    def test_01_data_file(self):
        """Tests run_script w/ data file"""
        debug.trace(4, f"TestIt.test_01_data_file(); self={self}")
        output = self.run_script(options="", env_options="", skip_stdin=True)
        # note: make sure some time mentioned for numpy
        # ex: numpy	0.304ms
        self.do_assert(my_re.search(r"numpy.*\d+.\d+\s*ms", output.strip()))
        return

    @pytest.mark.xfail                   # TODO: remove xfail
    ## DEBUG: @trap_exception            # TODO: remove when debugged
    def test_02_something_else(self):
        """Test for something_else: TODO..."""
        debug.trace(4, f"TestIt.test_02_something_else(); self={self}")
        self.do_assert(False)
        ## ex: self.do_assert(THE_MODULE.TODO_function() == TODO_value)
        return

#------------------------------------------------------------------------

if __name__ == '__main__':
    debug.trace_current_context()
    ## TODO2: here and elsewhere: invoke_tests(__file__)
    pytest.main([__file__])



#! /usr/bin/env python3
#
# Test(s) for Hugging Face (HF) Stable Diffulion (SD) module: ../hf_stable_diffusion.py
#
# Notes:
# - This can be run as follows:
#   $ PYTHONPATH=".:$PYTHONPATH" python ./mezcla/tests/test_hf_stable_diffusion.py
#
# TODO3: check image attributes (e.g., backgfround color) and work in image classification (e.g., box)
#
#--------------------------------------------------------------------------------
# Note:
# - This only runs if NVidia GPU support is installed OK, as determined via nvidia-smi output.
#   $ nvidia-smi
#   Fri Jan 26 12:51:57 2024       
#   +---------------------------------------------------------------------------------------+
#   | NVIDIA-SMI 535.98                 Driver Version: 535.98       CUDA Version: 12.2     |
#   ...
#   |   0  NVIDIA GeForce RTX 3080        Off | 00000000:01:00.0 Off |                  N/A |
#   |  0%   42C    P8              28W / 320W |    677MiB / 10240MiB |      0%      Default |
#   ...
#

"""Tests for hf_stable_diffusion module"""

# Standard packages
## OLD: import base64

# Installed packages
import numpy as np
import pytest
try:
    import diffusers
except:
    diffusers = None
try:
    import extcolors
except:
    extcolors = None
import PIL
import torch

# Local packages
from mezcla.unittest_wrapper import TestWrapper, trap_exception
from mezcla import debug
from mezcla import system
from mezcla import glue_helpers as gh
from mezcla.misc_utils import RANDOM_SEED
from mezcla.my_regex import my_re

# HACK: Makes sure low-resource GPU configuration used for tests
system.setenv("LOW_MEMORY", "1")
system.setenv("STREAMLINED_CLIP", "1")

# Note: Two references are used for the module to be tested:
#    THE_MODULE:                  global module object
#    TestTemplate.script_module:  path to file
import mezcla.examples.hf_stable_diffusion as THE_MODULE
hfsd = THE_MODULE

# Constants, etc.
TORCH_SEED = system.getenv_int("TORCH_SEED", RANDOM_SEED,
                               "Random seed for torch if non-zero")
NUMPY_SEED = system.getenv_int("NUMPY_SEED", RANDOM_SEED,
                               "Random seed for numpy if non-zero")
#................................................................................
# Utility functions

def get_gpu_stats():
    """
    Note: Currently uses NVIDIA System Management Interface (SMI) program"""
    # TODO3: use torch.cuda.memory_summary
    result = gh.run("nvidia-smi")
    debug.trace(7, f"get_gpu_stats() => {result}")
    return result


def gpu_mem_usage():
    """Return GPU memory usage in gigabytes"""
    # TEMP: Uses nvidia-smi until torch bug fixed
    # ex: |  0%   42C    P8              28W / 320W |    677MiB / 10240MiB |      0%      Default |
    total_usage = 0.0
    for line in get_gpu_stats().splitlines():
        if my_re.search(r"(\d+)MiB */ *(\d+)MiB", line, flags=my_re.IGNORECASE):
            device_usage, _device_total = my_re.groups()
            total_usage += system.to_float(device_usage)
    total_usage /= 1024
    debug.trace(6, f"get_gpu_stats() => {total_usage}")
    return total_usage

#................................................................................
# Globals

## TODO3: rework NVidia test via torch CUDA support
nvidia_ok = my_re.search("NVIDIA-SMI.*Driver Version",
                         get_gpu_stats(), flags=my_re.IGNORECASE)
sd_or_nvidia_issue = (not (diffusers and nvidia_ok))

#................................................................................

class TestIt(TestWrapper):
    """Class for testcase definition"""
    script_module = TestWrapper.get_testing_module_name(__file__, THE_MODULE)
    use_temp_base_dir = True            # treat TEMP_BASE as directory
    # note: temp_file defined by parent (along with script_module, temp_base, and test_num)
    init_gpu_mem_usage = gpu_mem_usage()

    @classmethod
    def setUpClass(cls, filename=None, module=None):
        """One-time initialization (i.e., for entire class)"""
        debug.trace(5, f"TestIt.setUpClass(); cls={cls}")
        # note: should do parent processing first
        super().setUpClass(filename, module)

        # Initialize torch's seed and numpy (n.b., separate from python, which misc_utils handles).
        # See https://pytorch.org/docs/stable/notes/randomness.html
        if TORCH_SEED:
            torch.manual_seed(TORCH_SEED)
        if NUMPY_SEED:
            np.random.seed(NUMPY_SEED)
        return

    def check_images(self, image_specs, label=None):
        """Make sure each of the IMAGE_SPECS are valid and return PIL image"""
        debug.trace(4, f"check_images({gh.elide(image_specs)}, {label})")
        label_spec = ("" if label is None else f"-{label}")
        images = []
        for i, spec in enumerate(image_specs):
            image = None
            try:
                # Make sure valid image
                image_bytes = hfsd.decode_base64_image(spec)
                image = hfsd.create_image(image_bytes)
                self.do_assert(isinstance(image, PIL.Image.Image))
            except:
                self.do_assert(False, "Problem decoding image spec")
            if image:
                images.append(image)
            # Save to disk if debugging
            if debug.debugging():
                temp_image_file = f"{self.temp_file}{label_spec}-{i + 1}.png"
                debug.trace_expr(4, temp_image_file)
                hfsd.write_image_file(temp_image_file, spec)
        return images

    
    @pytest.mark.skipif(sd_or_nvidia_issue, reason="SD or NVidia not setup")
    @pytest.mark.xfail                   # TODO: remove xfail
    def test_00_init_stable_diffusion(self):
        """Make sure ST module gets initialized OK"""
        debug.trace(4, f"test_00_init_stable_diffusion(); self={self}")
        ## HACK: This also serves to initialize the module for other tests (TODO2: fix init)
        THE_MODULE.init_stable_diffusion()
        self.do_assert(THE_MODULE.torch is not None)
        self.do_assert(THE_MODULE.sd_instance is not None)
        self.do_assert(isinstance(THE_MODULE.sd_instance, THE_MODULE.StableDiffusion))
        self.do_assert(self.init_gpu_mem_usage > 0)
        return
    
    @pytest.mark.skipif(sd_or_nvidia_issue, reason="SD or NVidia not setup")
    def test_01_simple_generation(self):
        """Makes sure simple image generation (txt2img) works as expected"""
        debug.trace(4, f"test_01_simple_generation(); self={self}")

        # Run script to generate orange ball and get image filename.
        # ex: "See sd-app-image-1.png for output image(s)."
        script_output = self.run_script(
            options="--batch --prompt 'orange ball' --negative 'green blue red yellow purple pink brown'",
            env_options=f"BASENAME='{self.temp_base}' LOW_MEMORY=1", uses_stdin=False)
        debug.trace_expr(5, script_output)
        self.do_assert(my_re.search(r"See (\S+.png) for output image\(s\).", script_output.strip()))
        image_file = my_re.group(1)

        # Make sure expected dimension and  imagetype
        # ex: sd-app-image-3.png: PNG image data, 512 x 512, 8-bit/color RGB, non-interlaced
        file_info = gh.run(f"file {image_file}")
        debug.trace_expr(5, file_info)
        self.do_assert(my_re.search("PNG image data, 512 x 512, 8-bit/color RGB", file_info))
        # TODO2: orange in rgb-color profile for image
        # TODO3: clip interrogator mentions "ball"
        return

    ## TODO?
    ## #...............................................................................
    ##
    ## class TestIt2:
    ##    """Another class for testcase definition
    ##    Note: Needed to avoid error with pytest due to inheritance with unittest.TestCase via TestWrapper"""
    
    @pytest.mark.skipif(sd_or_nvidia_issue, reason="SD or NVidia not setup")
    def test_02_txt2img_pipeline(self):
        """Make sure valid SD pipeline created"""
        debug.trace(4, f"test_02_txt2img_pipeline(); self={self}")
        sd = THE_MODULE.StableDiffusion(use_hf_api=True)
        pipe = sd.init_pipeline()
        actual = my_re.split(r"\W+", str(pipe))
        expect = "CLIPImageProcessor CLIPTextModel StableDiffusionSafetyChecker text_encoder".split()
        debug.trace_expr(5, actual, expect, delim="\n")
        self.do_assert(len(system.intersection(actual, expect)) > 2)
        return
    
    @pytest.mark.xfail                   # TODO: remove xfail
    @pytest.mark.skipif(sd_or_nvidia_issue, reason="SD or NVidia not setup")
    def test_03_txt2img_generation(self):
        """Make sure text-to-image reasonable"""
        debug.trace(4, f"test_03_txt2img_generation(); self={self}")
        sd = THE_MODULE.StableDiffusion(use_hf_api=True)
        ## TODO: NUM_IMAGES = 2
        NUM_IMAGES = 1
        # note: generate image with high adherence guidance for prompt
        image_specs = sd.infer(prompt="cute puppy", negative_prompt="pitbull", scale=20, num_images=NUM_IMAGES, skip_img_spec=True)
        images = self.check_images(image_specs, label="txt2img")
        self.do_assert(len(images) == NUM_IMAGES)
        description = sd.infer_img2txt(image_b64=image_specs[0])
        self.do_assert(my_re.search(r"canine|dog|puppy", description))
        return

    @pytest.mark.xfail                   # TODO: remove xfail
    @pytest.mark.skipif(sd_or_nvidia_issue, reason="SD or NVidia not setup")
    @trap_exception
    def test_04_img2img_generation(self):
        """Make sure image-to-image reasonable"""
        debug.trace(4, f"test_04_img2img_generation(); self={self}")
        sd = THE_MODULE.StableDiffusion(use_hf_api=True)
        ## TODO: NUM_IMAGES = 2
        NUM_IMAGES = 1
        # TODO2: use larger image
        PACMAC_LIKE_IMAGE = gh.resolve_path("sd-spooky-pacman.png", heuristic=True)
        pacmac_like_base64 = THE_MODULE.encode_image_file(PACMAC_LIKE_IMAGE)
        # note: generate derived image with high fidelity to original and low adherence guidance to prompt
        image_specs = sd.infer_img2img(image_b64=pacmac_like_base64, denoise=0.25, prompt="cute puppy", negative_prompt="pitbull",
                                       scale=3.5, num_images=NUM_IMAGES, skip_img_spec=True)
        self.do_assert(len(image_specs) == NUM_IMAGES)
        ## BAD: self.do_assert(isinstance(images[0], PIL.Image.Image))
        # TODO1: image recognition doesn't yield dog
        images = self.check_images(image_specs, label="img2img")
        self.do_assert(len(images) == NUM_IMAGES)
        description = sd.infer_img2txt(image_b64=image_specs[0])
        self.do_assert(not my_re.search(r"canine|dog|puppy", description))
        return

    @pytest.mark.xfail                   # TODO: remove xfail
    @pytest.mark.skipif(sd_or_nvidia_issue, reason="SD or NVidia not setup")
    @trap_exception
    def test_05_img2txt_generation(self):
        """Make sure image-to-text reasonable"""
        debug.trace(4, f"test_05_img2txt_generation(); self={self}")
        # TODO2: use common setup method (e.g., via TestWrapper)
        sd = THE_MODULE.StableDiffusion(use_hf_api=True, low_memory=True)
        PACMAC_LIKE_IMAGE = gh.resolve_path("sd-spooky-pacman.png", heuristic=True)
        pacmac_like_base64 = THE_MODULE.encode_image_file(PACMAC_LIKE_IMAGE)
        description = sd.infer_img2txt(image_b64=pacmac_like_base64)
        debug.trace_expr(5, description)
        # TODO4: assert(english-like-text(description))
        self.do_assert(not my_re.search(r"canine|dog|puppy", description))
        debug.trace(5, "post-test_05 GPU stats:\n" + get_gpu_stats())
        return

    @pytest.mark.xfail                   # TODO: remove xfail
    @pytest.mark.skipif(not extcolors, reason="extcolors package missing")
    @trap_exception
    def test_06_prompted_color(self):
        """Make sure prompted color is used in txt2img"""
        debug.trace(4, f"test_06_prompted_color(); self={self}")
        sd = THE_MODULE.StableDiffusion(use_hf_api=True, low_memory=True)
        images = sd.infer(prompt="a ripe orange", scale=30, skip_img_spec=True)
        # note: encodes image base-64 str data into bytes and then decodes into image bytes
        ## OLD: image_data = (base64.decodebytes(images[0].encode()))
        image_data = hfsd.decode_base64_image(images[0])
        image_path = gh.create_temp_file(image_data, binary=True)
        # note: use of rgb_color_name.py allows for fudge factor
        # $ extcolors sd-app-image-1.png | rgb_color_name.py - | grep orange
        # <(255, 92, 0), orangered>   :  47.07% (123388)
        # <(255, 153, 0), orange>  :   6.13% (16074)
        color_output = gh.run(f"extcolors '{image_path}' | rgb_color_name.py - 2> /dev/null")
        debug.trace_expr(4, color_output)
        self.do_assert("orange" in color_output)
        self.do_assert(len(images) == 1)
        return

    @pytest.mark.xfail                   # TODO: remove xfail
    @pytest.mark.skipif(not extcolors, reason="extcolors package missing")
    def test_99_gpu_mem_usage(self):
        """Ensure at least 2gb of GPU memory used
        Note: This should be the last test run
        """
        usage_before = self.init_gpu_mem_usage
        usage_after = gpu_mem_usage()
        usage = (usage_after - usage_before)
        debug.trace_expr(5, usage, usage_before, usage_after)
        self.do_assert(usage >= 2)
    
#------------------------------------------------------------------------

if __name__ == '__main__':
    debug.trace_current_context()
    try:
        pytest.main([__file__])
    except:
        system.print_exception_info("pytest.main")
    ## TEMP (Show GPU memory usage, etc.):
    ## NOTE: maldito pytest makes debugging a real pain!
    gpu_stats = "post test-suite GPU stats: {{\n{out}\n}}".format(
        out=gh.indent_lines(get_gpu_stats()))
    print(gpu_stats)
    debug.trace(3, gpu_stats)



#! /usr/bin/env python3
#
# Test(s) for ../bash_ast.py
#
# TODO:
# - Test for features in https://github.com/tomasohara/bashlex fork,
#   which attempts to make bashlex more usable in the wild.
#
#--------------------------------------------------------------------------------
# Sample output tested:
#
# $ bash_ast.py - <<<'let sum=2 + 2; if [ $sum -eq 5 ]; then echo "WTH?"; fi'
# ListNode(pos=(0, 54), parts=[
#   CommandNode(pos=(0, 13), parts=[
#     WordNode(pos=(0, 3), word='let'),
#     WordNode(pos=(4, 9), word='sum=2'),
#     WordNode(pos=(10, 11), word='+'),
#     WordNode(pos=(12, 13), word='2'),
#   ]),
#   OperatorNode(op=';', pos=(13, 14)),
#   CompoundNode(list=[
#     IfNode(pos=(15, 54), parts=[
# ...
#       ReservedwordNode(pos=(34, 38), word='then'),
#       ListNode(pos=(39, 51), parts=[
#           CommandNode(pos=(39, 50), parts=[
#             WordNode(pos=(39, 43), word='echo'),
#             WordNode(pos=(44, 50), word='WTH?'),
#           ]),
# ...
#   ], pos=(15, 54)),
# ])
#

# $ bash_ast.py - <<<'touch /tmp-$$/fubar; echo $?'
# ListNode(pos=(0, 28), parts=[
#   CommandNode(pos=(0, 19), parts=[
#     WordNode(pos=(0, 5), word='touch'),
#     WordNode(pos=(6, 19), word='/tmp-$$/fubar', parts=[
# ...
#   CommandNode(pos=(21, 28), parts=[
#     WordNode(pos=(21, 25), word='echo'),
#     WordNode(pos=(26, 28), word='$?', parts=[
# ...
# ])


"""Tests for bash_ast module"""

# Standard modules
## TODO: from collections import defaultdict

# Installed modules
import pytest

# Local modules
from mezcla.unittest_wrapper import TestWrapper, invoke_tests
from mezcla import debug
from mezcla.my_regex import my_re
from mezcla import system

# Note: Two references are used for the module to be tested:
#    THE_MODULE:                        global module object
#    TestIt.script_module:              path to file
try:
    ## TEST: raise RuntimeError
    import mezcla.bash_ast as THE_MODULE
except:
    THE_MODULE = None
    debug.trace_exception(4, "bash_ast import")

#------------------------------------------------------------------------

@pytest.mark.skipif(not THE_MODULE, reason="Unable to load module")
class TestIt(TestWrapper):
    """Class for command-line based testcase definition"""
    script_module = TestWrapper.get_testing_module_name(__file__, THE_MODULE)

    @pytest.mark.xfail                   # TODO: remove xfail
    def test_01_data_file(self):
        """Tests run_script w/ data file"""
        debug.trace(4, f"TestIt.test_01_data_file(); self={self}")
        data = ["let sum=2 + 2;",
                "if [ $sum -eq 5 ]; then echo 'WTH?'; fi"]
        system.write_lines(self.temp_file, data)
        output = self.run_script(options="", data_file=self.temp_file)
        self.do_assert(my_re.search(
            r"ListNode.*WordNode.*let.*ReservedwordNode.*if.*ReservedwordNode.*then",
            output, flags=my_re.DOTALL|my_re.MULTILINE))
        return

    @pytest.mark.xfail                   # TODO: remove xfail
    def test_02_something_else(self):
        """Test for something_else: TODO..."""
        debug.trace(4, f"TestIt.test_02_something_else(); self={self}")
        ast = THE_MODULE.BashAST("touch /tmp-$$/fubar; echo $?")
        output = ast.dump()
        self.do_assert(my_re.search(
            r"ListNode.*CommandNode.*touch.*CommandNode.*WordNode.*echo",
            output, flags=my_re.DOTALL|my_re.MULTILINE))
        return

#------------------------------------------------------------------------

if __name__ == '__main__':
    debug.trace_current_context()
    invoke_tests(__file__)



#! /usr/bin/env python3
#
# Tests for audio.py
#
# Notes:
# - *** All the tests are skipped if the audio packages not installed.
# - This can be run as follows:
#   $ PYTHONPATH=".:$PYTHONPATH" python ./mezcla/tests/test_audio.py
# - some tests will take a while.
#
# TODO2: remove posix dependencies (e.g., path)
#
# TODO3: solve test execution long times.
#
# TODO4:
# - Reword "this will take a while and this require a valid audio path in AUDIOFILE"
#   so that just covers RUN_SLOW_TESTS; Check separately for path.
#

"""Tests for audio module"""

# Standard modules
## NOTE: this is empty for now

# Installed modules
# note: The audio library modules are not installed by default, so tests skipped if not found.
import pytest
try:
    import librosa
except:
    librosa = None

# Local modules
from mezcla.unittest_wrapper import TestWrapper, invoke_tests
from mezcla import glue_helpers as gh
from mezcla import debug
from mezcla import system

# Note: Two references are used for the module to be tested:
#    THE_MODULE:	       module object (e.g., <module 'mezcla.main' ...>)
#    TestIt.script_module:     dotted module path (e.g., "mezcla.main")
if librosa:
    import mezcla.audio as THE_MODULE
else:
    debug.trace(3, "Unable to import librosa, so tests will be disabled")
    THE_MODULE = None

# Constants
# note: heuristic added for likely file move (e.g., ../examples => ../tests/resources)
RUN_SLOW_TESTS = system.getenv_bool(
    "RUN_SLOW_TESTS", False,
    ## TODO2: enable if running Github runner
    description="Run tests that can a while to run")
AUDIO_FILENAME = "fuzzy-testing-1-2-3.wav"
## OLD:
## AUDIOFILE = gh.resolve_path(AUDIO_FILENAME,
##                             base_dir=gh.form_path("..", "examples"))
AUDIOFILE = gh.resolve_path(gh.form_path("examples", AUDIO_FILENAME))
if system.file_exists(AUDIOFILE):
    AUDIOFILE = gh.resolve_path(AUDIO_FILENAME, base_dir=".", heuristic=True)
AUDIOFILE_TEXT = "testing one two three testing"
NOT_RUN_SLOW_REASON = f"this will take a while and/or a valid audio path ({AUDIOFILE})"

class TestAudio(TestWrapper):
    """Class for testcase definition"""
    script_module     = TestWrapper.get_testing_module_name(__file__, THE_MODULE)
    use_temp_base_dir = True
    maxDiff           = None

    @pytest.mark.skipif(not RUN_SLOW_TESTS, reason=NOT_RUN_SLOW_REASON)
    def test_sphinx_engine(self):
        """Test CMUSphinx speech recognition engine class"""
        debug.trace(debug.DETAILED, f"TestAudio.test_sphinx_engine({self})")
        sample = THE_MODULE.Audio(AUDIOFILE)
        result_speech = sample.speech_to_text(engine='sphinx')
        assert isinstance(result_speech, str)
        assert result_speech

    @pytest.mark.skipif(not librosa, reason="librosa missing")
    def test_audio_path(self):
        """Test for audio.path"""
        debug.trace(debug.DETAILED, f"TestAudio.test_audio_path({self})")
        ## TODO2: fixme
        ## OLD: path = 'some/path'
        path = gh.form_path("some", "path")
        #
        sample = THE_MODULE.Audio(path)
        assert sample.get_path() == path
        #
        sample = THE_MODULE.Audio()
        assert not sample.get_path()
        sample.set_path(path)
        assert sample.get_path(), path

    @pytest.mark.xfail                   # TODO: remove xfail
    def test_source_single_audio(self):
        """End to end test sourcing a single audio file"""
        debug.trace(debug.DETAILED, f"TestAudio.test_source_single_audio({self})")
        #
        audio = self.temp_file + '.wav'
        system.write_file(audio, 'some content')
        #
        ## BAD: command  = f'python {self.script_module} --verbose {audio}'
        actual = self.run_script(options="--verbose", data_file=audio)
        expected = f'Audio: {audio}'
        ## OLD: assert gh.run(command) == expected
        assert actual == expected

    @pytest.mark.xfail                   # TODO: remove xfail
    def test_source_list(self):
        """End to end test sourcing a list of audio files"""
        debug.trace(debug.DETAILED, f"TestAudio.test_source_list({self})")

        list_file = self.temp_file + '.list'

        audio_list = ('audio1.wav\n'
                      'audio2.wav\n'
                      'audio23.wav\n'
                      'audioN.wav')

        system.write_file(list_file, audio_list)

        ## BAD: command  = f'python {self.script_module} --verbose {list_file}'
        actual = self.run_script(options="--verbose", data_file=list_file)
        expected = ('Audio: audio1.wav\n'
                    'Audio: audio2.wav\n'
                    'Audio: audio23.wav\n'
                    'Audio: audioN.wav')
        ## OLD: assert gh.run(command) == expected
        assert actual == expected

    @pytest.mark.xfail                   # TODO: remove xfail
    def test_source_folder(self):
        """End to end test sourcing a folder and discover audiofiles"""
        debug.trace(debug.DETAILED, f"TestAudio.test_source_folder({self})")

        ## TODO: resolve librosa find files recursively not woking.
        ## filenames = ['/audio1.wav', '/somedir/audio2.wav', '/audio23.wav']

        filenames = ['/audio1.wav', '/audio23.wav']
        for filename in filenames:
            system.write_file(self.temp_base + filename, 'content')

        ## BAD: actual = gh.run(f'python {self.script_module} --verbose {self.temp_base}')
        actual = self.run_script(options="--verbose", data_file=self.temp_base)

        for filename in filenames:
            assert filename in actual

    @pytest.mark.skipif(not RUN_SLOW_TESTS, reason=NOT_RUN_SLOW_REASON)
    def test_extract_speech(self):
        """End to end test extracting speech using CMUSphinx engine"""
        debug.trace(debug.DETAILED, f"TestAudio.test_extract_speech({self})")

        ## BAD: actual = gh.run(f'python {self.script_module} --verbose --speech sphinx {AUDIOFILE}')
        actual = self.run_script(options="--verbose --speech sphinx ", data_file=AUDIOFILE)
        assert actual
        assert '- speech recognized' in actual
        assert isinstance(actual, str)

    @pytest.mark.skipif(not RUN_SLOW_TESTS, reason=NOT_RUN_SLOW_REASON)
    @pytest.mark.xfail                   # TODO: remove xfail
    def test_huggingface_asr(self):
        """Test Huggingface automatic speech recognition model class"""
        # NEW: Added test for Huggingface
        debug.trace(debug.DETAILED, f"TestAudio.test_huggingface_asr({self})")

        sample = THE_MODULE.Audio(AUDIOFILE)
        result_speech = sample.speech_to_text(engine='huggingface')
        print(result_speech)
        ## BAD: The result might differ according to the language models
        # result_speech_expected = "testing one two three testing"
        # assert (result_speech == result_speech_expected) -> AssertionError
        assert isinstance(result_speech, str)
        ## OLD: assert result_speech
        actual = result_speech.lower().split()
        expect = AUDIOFILE_TEXT.lower().split()
        debug.trace_expr(5, actual, expect)
        assert system.intersection(actual, expect)

if __name__ == '__main__':
    debug.trace_current_context()
    invoke_tests(__file__)



#! /usr/bin/env python3
#
# Note: Test for working around ":double import" issue. See
#    See https://stackoverflow.com/questions/43393764/python-3-6-project-structure-leads-to-runtimewarning
# Also see
#    https://stackoverflow.com/questions/4042905/what-is-main-py
#

"""Entry point for mezcla"""

# Standard module(s)
## TODO: import logging

# Installed modules
import mezcla
from mezcla import debug
from mezcla import glue_helpers as gh
from mezcla.my_regex import my_re
from mezcla import system

# Contants
TL = debug.TL

def main(omit_warnings=False):
    """Entry point: shows version info if debugging otherwise a warning"""
    if not omit_warnings:
        system.print_error(f"Warning: {__file__} is not intended as standalone.")
    file_path = system.real_path(gh.dirname(__file__))
    debug.trace(TL.USUAL, f"Version: {mezcla.__VERSION__}")
    debug.trace(TL.USUAL, f"Install path: {file_path}")

    # Derive module name
    # example: /home/tomohara/python/Mezcla/mezcla/__main__.py => "mezcla"
    module = "<module>"
    sep = my_re.escape(system.path_separator())
    match = my_re.search(fr"([^{sep}]*){sep}[^{sep}]*$", __file__)
    if match:
        module = match.group(1)
    if not omit_warnings:
        system.print_error(f"Likewise for the package (e.g., via 'python -m {module}')\n")

if __name__ == '__main__':
    main()



#! /usr/bin/env python3
#
# Test(s) for ../extract_document_text.py
#
# Notes:
# - For debugging the tested script, the ALLOW_SUBCOMMAND_TRACING environment
#   option shows tracing output normally suppressed by  unittest_wrapper.py.
# - This can be run as follows:
#   $ PYTHONPATH=".:$PYTHONPATH" python ./mezcla/tests/test_extract_document_text.py
#

"""Tests for extract_document_text module"""

# Standard packages
## OLD: import re

# Installed packages
import pytest

# Local packages
from mezcla import debug
from mezcla import glue_helpers as gh
from mezcla import system
from mezcla.unittest_wrapper import TestWrapper, invoke_tests


# Note: Two references are used for the module to be tested:
#    THE_MODULE:                        global module object
#    TestIt.script_module:              path to file
import mezcla.extract_document_text as THE_MODULE

class TestExtractDocumentText(TestWrapper):
    """Class for testcase definition"""
    script_module = TestWrapper.get_testing_module_name(__file__, THE_MODULE)
    
    @pytest.mark.xfail                   # TODO: remove xfail
    def test_document_to_text(self):
        """Ensure document_to_text() works as expected"""
        debug.trace(4, "test_document_to_text()")
        # NOTE: Output contains \n specifier at the end on the text.
        # - Uses test-specific temp file as per TestWrapper.
        # - Uses file extension for use by underlying textract module (see document_to_text).
        ## OLD: tmp_document = self.get_temp_file()
        tmp_document = self.temp_file + ".txt"

        # DOCUMENT_TEXT = """Lorem Ipsum is simply dummy text of the printing and typesetting industry.
        # Lorem Ipsum has been the industry's standard dummy text ever since the 1500s, when an unknown printer took a galley of type and scrambled it to make a type specimen book.
        # It has survived not only five centuries, but also the leap into electronic typesetting, remaining essentially unchanged.
        # It was popularised in the 1960s with the release of Letraset sheets containing Lorem Ipsum passages, and more recently with desktop publishing software like Aldus PageMaker including versions of Lorem Ipsum."""
        
        # Single-line string as INPUT
        DOCUMENT_TEXT = "Lorem ipsum dolor sit amet, consectetur adipiscing elit. Suspendisse porttitor rutrum leo. Nam efficitur justo quis mi ullamcorper consectetur. Aenean posuere nisl dignissim mauris blandit, sed euismod lacus posuere. Morbi ultrices magna eget faucibus ultricies. Praesent in ante nisi. Nulla faucibus sollicitudin sapien in aliquet. Interdum et malesuada fames ac ante ipsum primis in faucibus."
        
        system.write_file(tmp_document, text=DOCUMENT_TEXT)
        output = THE_MODULE.document_to_text(tmp_document)
        ## OLD:
        ## output_case1 = output != DOCUMENT_TEXT
        ## output_case2 = DOCUMENT_TEXT in output
        ## output_case3 = output.endswith("\n")
        ## output_case4 = output.strip() == DOCUMENT_TEXT
        ## assert output_case1 and output_case2 and output_case3 and output_case4
        assert output != DOCUMENT_TEXT
        assert DOCUMENT_TEXT in output
        assert output.endswith("\n")
        assert output.strip() == DOCUMENT_TEXT
        return

    @pytest.mark.xfail                   # TODO: remove xfail
    def test_document_to_text_IO(self):
        """Ensure document_to_text_IO() works as expected"""
        # TODO: Adding tests to check if extract_document_text.py works properly
        debug.trace(4, "test_document_to_text_IO()")
        ## OLD: env_option = "STDOUT"
        # NOTE: Output contains \n specifier at the end on the text.
        ## OLD:
        ## doc_input_temp = self.get_temp_file()
        ## doc_output_temp = self.get_temp_file()
        ## TODO2: carefully review tests/template.py
        doc_input_temp = self.temp_file + ".txt"
        # Single-line string as INPUT
        DOCUMENT_TEXT = "Lorem ipsum dolor sit amet, consectetur adipiscing elit. Suspendisse porttitor rutrum leo. Nam efficitur justo quis mi ullamcorper consectetur. Aenean posuere nisl dignissim mauris blandit, sed euismod lacus posuere. Morbi ultrices magna eget faucibus ultricies. Praesent in ante nisi. Nulla faucibus sollicitudin sapien in aliquet. Interdum et malesuada fames ac ante ipsum primis in faucibus."
        system.write_file(filename=doc_input_temp, text=DOCUMENT_TEXT)
        ## BAD:
        ## test_command_1 = f"../extract_document_text.py {doc_input_temp} {doc_output_temp}"
        ## gh.run(test_command_1)
        ## output = gh.read_file(doc_output_temp)
        output = self.run_script(env_options="STDOUT=1", data_file=doc_input_temp)
        
        ## BAD: assert output == DOCUMENT_TEXT
        assert output.strip() == DOCUMENT_TEXT
        return 
        
    def test_document_to_text_html(self):
        """Ensure document_to_text_html() works as expected"""
        debug.trace(4, "test_document_to_text_html()")
        data = [
            "<HTML lang='en'>",
            "  <BODY>",
            "    <H3>Hey</H3>"
            "  </BODY>",
            "</HTML>",
            ]
        temp_file = self.temp_file + ".html"
        system.write_lines(temp_file, data)
        output = self.run_script(options="", env_options="STDOUT=1", data_file=temp_file)
        self.do_assert("Hey" == output.strip())
        return

if __name__ == '__main__':
    debug.trace_current_context()
    invoke_tests(__file__)



#! /usr/bin/env python3
#
# Test(s) for ../file_utils.py
#
# Notes:
# - This can be run as follows:
#   $ PYTHONPATH=".:$PYTHONPATH" python ./mezcla/tests/test_file_utils.py
#

"""Tests for file_utils module"""

# Standard packages
import json
import re
import os
import datetime

# Installed packages
import pytest

# Local packages
from mezcla.unittest_wrapper import TestWrapper, invoke_tests
from mezcla import debug
from mezcla import glue_helpers as gh
from mezcla.my_regex import my_re
from mezcla import system

# Note: Two references are used for the module to be tested:
#    THE_MODULE:	                global module object
#    TestTemplate.script_module:        path to file
import mezcla.file_utils as THE_MODULE

class TestFileUtils(TestWrapper):
    """Class for testcase definition"""
    script_module = TestWrapper.derive_tested_module_name(__file__)
    use_temp_base_dir = True

    @pytest.mark.xfail                   # TODO: remove xfail
    def test_get_directory_listing(self):
        """
        Tests for get_directory_listing(path          = 'PATH',
                                        recursive     = False,
                                        long          = False,
                                        readable_size = False,
                                        return_string = True,
                                        make_unicode  = False)
        """

        # Setup files and folders
        folders     = [self.temp_base, gh.form_path(self.temp_base, 'other_folder')]
        filenames   = ['analize.py', 'debug.cpp', 'main.txt', 'misc_utils.perl']

        for foldername in folders:
            ## TODO3: use mezcla wrappers (e.g., gh.full_mkdir)
            system.create_directory(foldername)
            for filename in filenames:
                system.write_file(gh.form_path(foldername, filename), '')

        # Run Test
        list_result = THE_MODULE.get_directory_listing(f'{self.temp_base}', recursive=True, long=True, return_string=True)

        for line in list_result:
            assert bool(re.search(r"[drwx-]+\s+\d+\s+\w+\s+\w+\s+\d+\s+\w+\s+\d+\s+\d\d:\d\d\s+[\w/]+", line))

    @pytest.mark.xfail                   # TODO: remove xfail
    def test_get_information(self):
        """Tests for def get_information(path,
                                         readable = False,
                                         return_string = False)
        """
        temp_file = self.get_temp_file()
        system.write_file(temp_file, '')

        if os.name == 'nt':
            ls_result = gh.run(f'dir {temp_file} /Q').split('\n')[5].strip()
            ls_result = re.sub(r'\s+', ' ', ls_result).split(' ')
            modif_time = ls_result[1]
            owner = ls_result[3].split('\\')[1]
            file_name = ls_result[-1]
            info = THE_MODULE.get_information(temp_file, return_string=False)
            
            assert modif_time in info[5]
            assert owner == info[2]
            assert file_name in info[-1]
            assert 'None' == info[3]
            assert os.stat(temp_file).st_size == int(info[4])
            # assert ls_result.lower() == info.lower()
        else:
            ls_result = gh.run(f'ls -l {temp_file}')
            ls_result = re.sub(r'\s+', ' ', ls_result)
            assert THE_MODULE.get_information(temp_file, return_string=True).lower() == ls_result.lower()

    @pytest.mark.xfail                   # TODO: remove xfail
    def test_get_permissions(self):
        """Tests for get_permissions(path)"""

        # Setup
        test_file = gh.form_path(self.temp_base, 'some-file.cpp')
        system.write_file(test_file, '')

        # Run
        ## OLD:
        ## assert THE_MODULE.get_permissions(test_file) == '-rw-rw-rw-'
        ## assert THE_MODULE.get_permissions(self.temp_base) == 'drwxrwxrwx'
        assert my_re.search("-rw-r..r..", THE_MODULE.get_permissions(test_file))
        assert my_re.search("drwxr..r..", THE_MODULE.get_permissions(self.temp_base))

    @pytest.mark.xfail                   # TODO: remove xfail
    def test_get_modification_date(self):
        """Tests for get_modification_date(path)"""
        system.write_file(self.temp_file, '')
        ls_date = os.stat(self.temp_file).st_mtime
        # Extract date
        char = '#' if os.name == 'nt' else '-'
        ls_date = datetime.datetime.fromtimestamp(ls_date).strftime(f'%b %{char}d %H:%M').lower()

        assert THE_MODULE.get_modification_date(self.temp_file).lower() == ls_date

    @pytest.mark.xfail                   # TODO: remove xfail
    def test_json_to_jsonl(self):
        """Tests for json_to_jsonl"""
        in_file = gh.form_path(self.temp_base, "some-file.json")
        out_file = gh.form_path(self.temp_base, "some-file.jsonl")
        sample_array = [1, 2, 3]
        system.write_file(in_file, f"{sample_array}\n")
        THE_MODULE.json_to_jsonl(in_file, out_file)
        assert(system.read_lines(out_file) == list(map(str, sample_array)))

    @pytest.mark.xfail                   # TODO: remove xfail
    def test_jsonl_to_json(self):
        """Tests for jsonl_to_json"""
        in_file = gh.form_path(self.temp_base, "another-file.jsonl")
        out_file = gh.form_path(self.temp_base, "another-file.json")
        sample_array = ["dog", "cat", {"fav": "???"}]
        system.write_lines(in_file, [json.dumps(item) for item in sample_array])
        THE_MODULE.jsonl_to_json(in_file, out_file)
        assert(json.loads(system.read_file(out_file)) == sample_array)


#------------------------------------------------------------------------

if __name__ == '__main__':
    debug.trace_current_context()
    invoke_tests(__file__)



# Script with good argument type for tests/test_validate_arguments.py

"""
Some docs here
"""

from mezcla import debug

def not_called_function() -> None:
    """This function is not called"""
    print("This function is not called")
    print(debug.get_level())

def move_debug_level() -> None:
    """Move debug level"""
    print('Default debug level: ', debug.DEFAULT)
    print('Current debug level: ', debug.get_level())
    debug.set_level(debug.get_level() + 1)
    debug.set_level(debug.get_level() - 1)

def main() -> None:
    """Main function"""
    print("Hello, ...")
    move_debug_level()
    print("... World!")

if __name__ == "__main__":
    main()



#!/usr/bin/env python3

## 2023-12-30 19:09 (+05:45 GMT)
## plot.py was created to get statistics about the time consumed on CPU instances (JarvisLabs.ai)

import matplotlib.pyplot as plt

USE_GPU = True
time_consumed_cpu = [39.2, 40.2, 40.3, 39.6, 39.5, 39.4, 40.1, 39.5, 37.9, 38.6, 38.8, 38.8, 38.9, 39.6, 39.6]
time_consumed_gpu = [28.212, 25.740, 27.163, 25.947, 27.393, 28.302, 26.352, 26.979, 27.7, 26.36, 29.95, 26.941, 27.5, 27.5, 27.046]
time_arr = time_consumed_gpu if USE_GPU else time_consumed_cpu

# # For boxplot
def box_plot(time_consumed, title:str="Time Consumed"):
    plt.boxplot(time_consumed)
    plt.xlabel('Test Number')
    plt.ylabel('Consumption (sec.)')
    plt.title(title)
    plt.show()

# For bar plot
def bar_plot(time_consumed, title:str="Time Consumed"):
    plt.bar(range(len(time_consumed)), time_consumed, color='skyblue')
    plt.xlabel('Test Number')
    plt.xticks(range(len(time_consumed)))
    plt.ylabel('Consumption (sec.)')
    plt.title(title)
    plt.grid(True, linestyle='--', alpha=0.7)
    plt.show()

bar_plot(time_arr, "Time Consumed by JarvisLabs (GPU): RTX A5000 16GB")


#! /usr/bin/env python3
#
# text_processing.py: performs text processing (e.g., via NLTK), mainly for
# word tokenization and part-of-speech tagging. To make the part-of-speech
# output easier to digrest, there is an envionment option USE_CLASS_TAGS
# to enable use of class-based tags, such as those used in traditional 
# grammar (e.g., noun for NNS).
# 
#
# Notes:
# - function resulting caching ("memoization") is used via memodict decoration
# - environment variables (env-var) are used for some adhoc options
# - to bypass NLTK (e.g., for quick debugging), set SKIP_NLTK env-var to 0
#
#------------------------------------------------------------------------
# Miscelleneous notes
#
# Penn Tags (used by NLTK part-of-speech tagger)
# 
# CC    Coordinating conjunction
# CD    Cardinal number
# DT    Determiner
# EX    Existential there
# FW    Foreign word
# IN    Preposition or subordinating conjunction
# JJ    Adjective
# JJR   Adjective, comparative
# JJS   Adjective, superlative
# LS    List item marker
# MD    Modal
# NN    Noun, singular or mass
# NNS   Noun, plural
# NNP   Proper noun, singular
# NNPS  Proper noun, plural
# PDT   Predeterminer
# POS   Possessive ending
# PP    Personal pronoun
# PP$   Possessive pronoun ???
# PRP$  Possessive pronoun
# PRP   Personal pronoun
# RB    Adverb
# RBR   Adverb, comparative
# RBS   Adverb, superlative
# RP    Particle
# SYM   Symbol
# TO    to
# UH    Interjection
# VB    Verb, base form
# VBD   Verb, past tense
# VBG   Verb, gerund or present participle
# VBN   Verb, past participle
# VBP   Verb, non-3rd person singular present
# VBZ   Verb, 3rd person singular present
# WDT   Wh-determiner
# WP    Wh-pronoun
# WP$   Possessive wh-pronoun
# WRB   Wh-adverb
#
# See ftp://ftp.cis.upenn.edu/pub/treebank/doc/tagguide.ps.gz for more details.
#------------------------------------------------------------------------
# TODO:
# - Integrate punctuation isolation (see prep_brill.perl).
# - Add usage examples (e.g., part-of-speech tagging).
# - Warn that without NLTK or Enchant, the text processing just checks common cases.
#

"""Performs text processing (e.g., tokenization via NLTK)"""

# Standard packages
from abc import ABCMeta, abstractmethod
import sys                              # system interface (e.g., command line)
import re                               # regular expressions

# Installed packages
## TODO: import flair
## OLD:
## from flair.data import Sentence
## from flair.models import SequenceTagger
flair = None

# Local packages
from mezcla import debug
## OLD: from mezcla import misc_utils
## OLD: from mezcla import spacy_nlp
spacy_nlp = None
from mezcla import system
from mezcla import glue_helpers as gh
## OLD: from mezcla import tpo_common as tpo

# Constants
TL = debug.TL

#------------------------------------------------------------------------
# Globals

# Object for spell checking via enchant
speller = None
word_freq_hash = None
WORD_FREQ_FILE = system.getenv_text("WORD_FREQ_FILE", "word.freq")
nltk = None
enchant = None

# Hash for returning most common part of speech for a word (or token)
word_POS_hash = None
WORD_POS_FREQ_FILE = system.getenv_text("WORD_POS_FREQ_FILE", "word-POS.freq")

# Misc. options
LINE_MODE = system.getenv_boolean("LINE_MODE", False, "Process text line by line (not all all once)")
JUST_TAGS = system.getenv_boolean("JUST_TAGS", False, "Just show part of speech tags (not word/POS pair)")
OUTPUT_DELIM = system.getenv_text("OUTPUT_DELIM", " ", "Delimiter (separator) for output lists")
INCLUDE_MISSPELLINGS = system.getenv_boolean("INCLUDE_MISSPELLINGS", False, "Include spell checking")
SKIP_MISSPELLINGS = system.getenv_boolean("SKIP_MISSPELLINGS", not INCLUDE_MISSPELLINGS, "Skip spell checking")
SHOW_MISSPELLINGS = not SKIP_MISSPELLINGS
USE_CLASS_TAGS = system.getenv_boolean("USE_CLASS_TAGS", False, "Use class-based tags (e.g., Noun for NNP)")
SKIP_CLASS_TAGS = not USE_CLASS_TAGS
KEEP_PUNCT = system.getenv_boolean("KEEP_PUNCT", False, "Use punctuation symbol as part-of-speech label")
TERSE_OUTPUT = system.getenv_boolean("TERSE_OUTPUT", JUST_TAGS, "Terse output mode")
VERBOSE = not TERSE_OUTPUT

# Skip use of NLTK and/or ENCHANT packages (using simple versions of functions)
# TODO: make misspellings optional (add --classic mode???)
SKIP_NLTK = system.getenv_boolean("SKIP_NLTK", False, "Omit usage of Natural Language Toolkit (NLTK)")
SKIP_ENCHANT = system.getenv_boolean("SKIP_ENCHANT", SKIP_MISSPELLINGS, "Omit usage of Enchant package for spell checking")
DOWNLOAD_DATA = system.getenv_bool("DOWNLOAD_DATA", False,
                                   "Download data from NLTK")
FLAIR_MODEL = system.getenv_text(
    "FLAIR_MODEL", "flair/chunk-english-fast",
    description="Model for Flair: see https://huggingface.co/flair"
)
TEXT_PROC = system.getenv_text(
    "TEXT_PROC", "spacy",
    description="name of text processor to use for chunking")

# List of stopwords (e.g., high-freqency function words)
stopwords = None

#------------------------------------------------------------------------
# Optional libraries
## TODO4: add init function for dynamic loading (e.g., for testing purposes)

# NLP toolkit
if not SKIP_NLTK:
    import nltk            # pylint: disable=ungrouped-imports
    if DOWNLOAD_DATA:
        download_nltk_resources()
# spell checking
if not SKIP_ENCHANT:
    import enchant         # pylint: disable=ungrouped-imports

#------------------------------------------------------------------------
# Functions

def split_sentences(text):
    """Splits TEXT into sentences"""
    # EX: split_sentences("I came. I saw. I conquered!") => ["I came.", "I saw.", "I conquered!"]
    # EX: split_sentences("Dr. Watson, it's elementary. But why?") => ["Dr. Watson, it's elementary.", "But why?"]
    if SKIP_NLTK:
        # Split around sentence-ending punctuation followed by space,
        # but excluding initials (TODO handle abbreviations (e.g., "mo.")
        #
        # TEST: Replace tabs with space and newlines with two spaces
        ## text = re.sub(r"\t", " ", text)
        ## text = re.sub(r"\n", "  ", text)
        # 
        # Make sure ending punctuaion followed by two spaces and preceded by one
        text = re.sub(r"([\.\!\?])\s", r" \1  ", text)
        #
        # Remove spacing added above after likely abbreviations
        text = re.sub(r"\b([A-Z][a-z]*)\s\.\s\s", r"\1. ", text)
        #
        # Split sentences by ending punctuation followed by two spaces
        # Note: uses a "positive lookbehind assertion" (i.e., (?<=...) to retain punctuation 
        sentences = re.split(r"(?<=[\.\!\?])\s\s+", text.strip())
    else:
        sentences = nltk.tokenize.sent_tokenize(text)
    return sentences


def split_word_tokens(text, omit_punct=False, omit_stop=None):
    """Splits TEXT into word tokens (i.e., words, punctuation, etc.), optionally with OMIT_PUNCT and OMIT_STOP.
    Note: run split_sentences first (e.g., to allow for proper handling of periods).
    By default, this uses NLTK's PunktSentenceTokenizer."""
    # EX: split_word_tokens("How now, brown cow?") => ['How', 'now', ',', 'brown', 'cow', '?']
    debug.trace(7, "split_word_tokens(%s); type=%s" % (text, type(text)))
    if SKIP_NLTK:
        tokens = [t.strip() for t in re.split(r"(\W+)", text) if (len(t.strip()) > 0)]
    else:
        tokens = nltk.word_tokenize(text)
    if omit_punct:
        tokens = [t for t in tokens if not is_punct(t)]
    if omit_stop:
        tokens = [t for t in tokens if not is_stopword(t)]
    debug.trace(7, "tokens: %s" % [(t, type(t)) for t in tokens])
    return tokens
#
# split_word_tokens("How now, brown cow?", omit_punct=True, omit_stop=True) => ['now', 'brown', 'cow']


def label_for_tag(POS, word=None):
    """Returns part-of-speech label for POS, optionally overriden based on WORD (e.g., original token if punctuation)"""
    label = POS
    if KEEP_PUNCT and word and re.match(r"\W", word[0]):
        label = word
        debug.trace_fmtd(5, "label_for_tag({t}, {w}) => {l}", t=POS, w=word, l=label)
    return label


fallback_POS_class = {  # classes for parts-of-speech not covered by rules (see class_for_tag)
    "CC": "conjunction",        # Coordinating conjunction
    "CD": "number",             # Cardinal number
    "DT": "determiner",         # Determiner
    "EX": "pronoun",            # Existential there
    "FW": "noun",               # Foreign word
    "IN": "preposition",        # Preposition or subordinating conjunction
    "JJ": "adjective",          # Adjective
    "JJR": "adjective",         # Adjective, comparative
    "JJS": "adjective",         # Adjective, superlative
    "LS": "punctuation",        # List item marker
    "MD": "auxiliary",          # Modal
    "PDT": "determiner",        # Predeterminer
    "POS": "punctuation",       # Possessive ending
    "SYM": "punctuation",       # Symbol
    "TO": "preposition",        # to
    "UH": "punctuation",        # Interjection
    "WDT": "determiner",        # Wh-determiner
    "WP": "pronoun",            # Wh-pronoun
    "WP$": "pronoun",           # Possessive wh-pronoun
    "WRB": "adverb",            # Wh-adverb
}


def class_for_tag(POS, word=None, previous=None):
    """Returns class label for POS tag, optionally over WORD and for PREVIOUS tag. Note: most cases are resolved without previous tag for context, except for gerunds and past participles which are considered nouns unless previous is auxiliary. Similarly, the word is only considered for special cases like punctuation."""
    # EX: class_for_tag("NNS") => "noun"
    # EX: class_for_tag("VBG") => "verb"
    # EX: class_for_tag("VBG", previous="MD") => "verb"
    # EX: class_for_tag("NNP", word="(") => "punctuation"
    tag_class = "unknown"
    if (POS == "VBG") and previous and ((previous[:2] == "VB") or (previous == "MD")):
        tag_class = "verb"
    elif (POS == "VBP") and previous and ((previous[:2] == "VB") or (previous == "MD")):
        tag_class = "verb"
    elif word and re.match(r"\W", word[0]):
        tag_class = "punctuation"
    # TODO: use prefix-based lookup table for rest (e.g., {"NN": "noun", ...})?
    elif POS[:2] == "NN":
        tag_class = "noun"
    elif POS[:2] == "JJ":
        tag_class = "adjective"
    elif POS[:2] == "VB":
        tag_class = "verb"
    elif POS[:2] == "RB":
        tag_class = "adverb"
    elif (POS[:2] == "PR") or (POS[:2] == "PP"):
        tag_class = "pronoun"
    else:
        tag_class = fallback_POS_class.get(POS, tag_class)
    debug.trace_fmtd(5, "class_for_tag({t}, {p}) => {c}", t=POS, p=previous, c=tag_class)
    return tag_class


def tag_part_of_speech(tokens):
    """Return list of part-of-speech taggings of form (token, tag) for list of TOKENS"""
    # EX: tag_part_of_speech(['How', 'now', ',', 'brown', 'cow', '?']) => [('How', 'WRB'), ('now', 'RB'), (',', ','), ('brown', 'JJ'), ('cow', 'NN'), ('?', '.')]
    if SKIP_NLTK:
        part_of_speech_taggings = [(word, get_most_common_POS(word)) for word in tokens]
    else:
        part_of_speech_taggings = []
        previous = None
        raw_part_of_speech_taggings = nltk.pos_tag(tokens)
        debug.trace(3, "raw tags: %s" % [t for (_w, t) in raw_part_of_speech_taggings])
        for (word, POS) in raw_part_of_speech_taggings:
            tag = label_for_tag(POS, word) if SKIP_CLASS_TAGS else class_for_tag(POS, word, previous)
            part_of_speech_taggings.append((word, tag))
            previous = POS
    debug.trace_fmt(6, "tag_part_of_speech({tokens}) => {tags}", 
                    tokens=tokens, tags=part_of_speech_taggings)
    return part_of_speech_taggings


def tokenize_and_tag(text):
    """Run sentence(s) and word tokenization over text and then part-of-speech tag it"""
    # TODO: return one list per sentence not just one combined list
    debug.trace(7, "tokenize_and_tag(%s)" % text)
    ## OLD: text = tpo.ensure_unicode(text)
    text_taggings = []
    for sentence in split_sentences(text):
        debug.trace(4, "sentence: %s" % sentence.strip())
        tokens = split_word_tokens(sentence)
        debug.trace(4, "tokens: %s" % tokens)
        taggings = tag_part_of_speech(tokens)
        debug.trace(4, "taggings: %s" % taggings)
        text_taggings += taggings
    return text_taggings


def tokenize_text(text):
    """Run sentence and word tokenization over text returning a list of sentence word lists"""
    tokenized_lines = []
    for sentence in split_sentences(text):
        sent_tokens = split_word_tokens(sentence)
        debug.trace(4, "sent_tokens: %s" % sent_tokens)
        tokenized_lines.append(sent_tokens)
    return tokenized_lines


@system.memodict
def is_stopword(word):
    """Indicates whether WORD should generally be excluded from analysis (e.g., function word)"""
    # note: Intended as a quick filter for excluding non-content words.
    global stopwords
    if (stopwords is None):
        if SKIP_NLTK:
            stopwords = ['i', 'me', 'my', 'myself', 'we', 'our', 'ours', 'ourselves', 'you', 'your', 'yours', 'yourself', 'yourselves', 'he', 'him', 'his', 'himself', 'she', 'her', 'hers', 'herself', 'it', 'its', 'itself', 'they', 'them', 'their', 'theirs', 'themselves', 'what', 'which', 'who', 'whom', 'this', 'that', 'these', 'those', 'am', 'is', 'are', 'was', 'were', 'be', 'been', 'being', 'have', 'has', 'had', 'having', 'do', 'does', 'did', 'doing', 'a', 'an', 'the', 'and', 'but', 'if', 'or', 'because', 'as', 'until', 'while', 'of', 'at', 'by', 'for', 'with', 'about', 'against', 'between', 'into', 'through', 'during', 'before', 'after', 'above', 'below', 'to', 'from', 'up', 'down', 'in', 'out', 'on', 'off', 'over', 'under', 'again', 'further', 'then', 'once', 'here', 'there', 'when', 'where', 'why', 'how', 'all', 'any', 'both', 'each', 'few', 'more', 'most', 'other', 'some', 'such', 'no', 'nor', 'not', 'only', 'own', 'same', 'so', 'than', 'too', 'very', 's', 't', 'can', 'will', 'just', 'don', 'should', 'now']
        else:
            stopwords = nltk.corpus.stopwords.words('english')
        debug.trace(4, "stopwords: %s" % stopwords)
    return (word.lower() in stopwords)


@system.memodict
def has_spelling_mistake(term):
    """Indicates whether TERM represents a spelling mistake"""
    # TODO: rework in terms of a class-based interface
    has_mistake = False
    try:
        if SKIP_ENCHANT:
            global word_freq_hash
            if not word_freq_hash:
                word_freq_path = gh.resolve_path(WORD_FREQ_FILE)
                gh.assertion(gh.non_empty_file(word_freq_path))
                word_freq_hash = read_freq_data(word_freq_path)
            has_mistake = term.lower() not in word_freq_hash
        else:
            global speller
            if not speller:
                speller = enchant.Dict("en_US")
            has_mistake = not speller.check(term)
    except:
        system.print_exception_info(f"spell checking of {term}")
    debug.trace(6, f"has_spelling_mistake({term}) => {has_mistake}")
    return has_mistake


def read_freq_data(filename):
    """Reads frequency listing for words (or other keys). A hash table is returned from (lowercased) key to their frequency."""
    # Sample input:
    #   # Word      Freq
    #   the 179062
    #   to  123567
    #   is  99390
    #   and 95920
    #   a   76679
    debug.trace(3, "read_freq_data(%s)" % filename)
    freq_hash = {}

    # Process each line of the file
    input_handle = system.open_file(filename)
    line_num = 0
    for line in input_handle:
        line_num += 1
        # Ignore comments
        line = line.strip()
        if (len(line) > 0) and (line[0] == '#'):
            continue

        # Extract the four fields and warn if not defined
        fields = line.split("\t")
        if (len(fields) != 2):
            debug.trace(3, "Ignoring line %d of %s: %s" % (line_num, filename, line))
            continue
        (key, freq) = fields
        key = key.strip().lower()

        # Store in hash
        if key not in freq_hash:
            freq_hash[key] = freq
        else:
            debug.trace(6, "Ignoring alternative freq for key %s: %s (using %s)" 
                        % (key, freq, freq_hash[key]))
        
    return (freq_hash)


def read_word_POS_data(filename):
    """Reads frequency listing for words in particular parts of speech
    to derive dictionay of the most common part-of-speech for words (for quick-n-dirty part-of-speech
    tagging). A hash table is returned from (lowercased) words to their most common part-of-speech."""
    # Sample input:
    #   # Token     POS     Freq
    #   ,           ,       379752
    #   .           .       372550
    #   the         DT      158317
    #   to          TO      122189
    debug.trace(3, "read_word_POS_freq(%s)" % filename)
    global word_POS_hash
    word_POS_hash = {}

    # Process each line of the file
    input_handle = system.open_file(filename)
    line_num = 0
    for line in input_handle:
        line_num += 1
        # Ignore comments
        line = line.strip()
        if (len(line) > 0) and (line[0] == '#'):
            continue

        # Extract the four fields and warn if not defined
        fields = line.split("\t")
        if (len(fields) != 3):
            debug.trace(3, "Ignoring line %d of %s: %s" % (line_num, filename, line))
            continue
        (word, POS, _freq) = fields
        word = word.strip().lower()

        # Store in hash
        if word not in word_POS_hash:
            word_POS_hash[word] = POS
        else:
            debug.trace(6, "Ignoring alternative POS for word %s: %s (using %s)" 
                            % (word, POS, word_POS_hash[word]))
        
    return (word_POS_hash)


def get_most_common_POS(word):
    """Returns the most common part-of-speech label for WORD, defaulting to NN (noun)"""
    # EX: get_most_common_POS("can") => "MD"
    # EX: get_most_common_POS("notaword") => "NN"
    global word_POS_hash
    if not word_POS_hash:
        word_POS_freq_path = gh.resolve_path(WORD_POS_FREQ_FILE)
        gh.assertion(gh.non_empty_file(word_POS_freq_path))
        word_POS_hash = read_word_POS_data(word_POS_freq_path)
    label = "NN"
    word = word.lower()
    if (word in word_POS_hash):
        label = word_POS_hash[word]
    return label

#------------------------------------------------------------------------
# Utility functions
#
# TODO: make POS optional for is_POS-type functions (and use corpus frequencies to guess)
#

def is_noun(_token, POS):
    """Indicates if TOKEN is a noun, based on POS"""
    return (POS[0:2] == "NN")


def is_verb(_token, POS):
    """Indicates if TOKEN is a verb, based on POS"""
    # EX: is_verb('can', 'NN') => False
    return (POS[0:2] == "VB")


def is_adverb(_token, POS):
    """Indicates if TOKEN is an adverb, based on POS"""
    # EX: is_adverb('quickly', 'RB') => True
    return (POS[0:2] == "RB")


def is_adjective(_token, POS):
    """Indicates if TOKEN is an adjective, based on POS"""
    # EX: is_adverb('quick', 'JJ') => True
    return (POS[0:2] == "JJ")


def is_comma(token, POS):
    """Indicates if TOKEN is a comma"""
    return ((token == ",") or (POS[0:1] == ","))


def is_quote(token, _POS):
    """Indicates if TOKEN is a quotation mark"""
    # Note: this includes checks for MS Word smart quotes because in training data
    # TODO: make handled properly with respect to Unicode encoding (e.g., UTF-8)
    return token in "\'\"\x91\x92\x93\x94"

def is_punct(token, POS=None):
    """Indicates if TOKEN is a punctuation symbol"""
    # EX: is_punct('$', '$') => True
    return (re.search("[^A-Za-z0-9]", token[0:1]) or 
            (POS and re.search("[^A-Za-z]", POS[0:1])))

# TODO: alternative to is_punct
## def is_punctuation(token, POS):
##     """Indicates if TOKEN is a punctuation symbol"""
##     # EX: is_punct('$', '$') => True
##     # TODO: find definitive source (or use ispunct-type function)
##     punctuation_chars_regex = r"[\`\~\!\@\#\$\%\^\&\*\(\)\_\-\+\=\{\}\[\]\:\;\"\'\<\>\,\.]"
##     return (re.search(punctuation_chars_regex, token[0:1]) or (POS
## re.search("[^A-Za-z]", POS[0:1]))


def init():
    """Makes sure globals defined"""

    # Note: stop word check is done for side of effect of loading list
    unexpected_stopword = is_stopword("movement")
    debug.assertion(not unexpected_stopword)

#-------------------------------------------------------------------------------

class TextProc(metaclass=ABCMeta):
    """Base class for text processing, such as via NLTK, Spacy or Flair"""

    def __init__(self, model=None):
        """Initializer: with provider-specific MODEL"""
        debug.trace(TL.VERBOSE, f"TextProc.__init__({model}): self={self}")
        debug.trace_object(5, self, label="TextProc instance")

    @abstractmethod
    def noun_phrases(self, text):
        """Return list of noun chunks in text"""
        raise NotImplementedError()
        
    @abstractmethod
    def verb_phrases(self, text):
        """Return list of verb chunks in text"""
        raise NotImplementedError()
        
class SpacyTextProc(TextProc):
    """TextProc via Spacy"""
    
    def __init__(self, model=None):
        super().__init__(model)
        global spacy_nlp
        if spacy_nlp is None:
            # pylint: disable=import-outside-toplevel, disable=redefined-outer-name
            from mezcla import spacy_nlp
        self.spacy = spacy_nlp.Chunker(model)
        debug.trace_object(5, self, label="SpacyTextProc instance")

    def noun_phrases(self, text):
        return self.spacy.noun_phrases(text)

    def verb_phrases(self, text):
        debug.trace(4, "FYI: Spacy doesn't support verb phrase chunking")
        result = []
        debug.trace(6, f"noun_phrases({text!r}) => {result!r}")
        return result

class FlairTextProc(TextProc):
    """TextProc via Flair"""

    def __init__(self, model=None):
        super().__init__(model)
        if model is None:
            model = FLAIR_MODEL
        global flair
        if flair is None:
            # pylint: disable=import-outside-toplevel, disable=redefined-outer-name
            import flair
        self.tagger = flair.models.SequenceTagger.load(model)
        debug.trace_object(5, self, label="FlairTextProc instance")
    
    def noun_phrases(self, text):
        # Based on https://huggingface.co/flair/chunk-english-fast
        sentence = flair.data.Sentence(text)
        self.tagger.predict(sentence)
        noun_phrases = [phr.text for phr in sentence.get_spans()  if (phr.tag == "NP")]
        debug.trace(6, f"noun_phrases({text!r}) => {noun_phrases!r}")
        return noun_phrases

    def verb_phrases(self, text):
        # Based on https://huggingface.co/flair/chunk-english-fast
        sentence = flair.Sentence(text)
        self.tagger.predict(sentence)
        verb_phrases = [phr.text for phr in sentence.get_spans()  if (phr.tag == "VP")]
        debug.trace(6, f"verb_phrases({text!r}) => {verb_phrases!r}")
        return verb_phrases


def create_text_proc(name, *args, **kwargs):
    """Return class to use for TextProc NAME
    Note: Passes along optional ARGS and KWARGS to constructor
    """
    # EX: (create_text_proc("flair").__class__ == FlairTextProc().__class__)
    # TODO2: see if standard way to do class displatching
    class_instance = None
    try:
        ## BAD: class_object = misc_utils.get_class_from_name(name.capitalize())
        ## TODO: class_object = misc_utils.get_class_from_name(name.capitalize(), module_name=__name__)
        ## OLD: class_object = (SpacyTextProc if (name == "spacy") else FlairTextProc if (name == "flair") else TextProc)
        class_object = (SpacyTextProc if (name == "spacy") else FlairTextProc if (name == "flair") else None)
        ## BAD:
        ## debug.assertion(class_object.__class__.__name__.endswith("TextProc"))
        ## class_instance = class_object.__new__(*args, **kwargs)
        class_instance = class_object(*args, **kwargs)
        debug.trace_expr(6, class_object, class_instance)
    except:
        system.print_exception_info("create_text_proc")
    return class_instance


def download_nltk_resources():
    """Download NLTK resources from their website"""
    nltk.download(['punkt', 'punkt_tab', 
                   'averaged_perceptron_tagger',
                   'averaged_perceptron_tagger_eng',
                   'stopwords'])

#-------------------------------------------------------------------------------

def usage():
    """Displays command-line usage"""
    usage_note = """
Usage: _SCRIPT_ [--help] [--just-tokenize] [--just-chunk] [--lowercase] file
Example:

echo "My dawg has fleas." | _SCRIPT_ -

Notes:
- Intended more as a library module
- In standalone mode it runs the text processing pipeline over the file:
     sentence splitting, word tokenization, and part-of-speech tagging
- Set SKIP_NLTK environment variable to 1 to disable NLTK usage.
"""
    # TODO: __file__ => sys.argv[1]???
    system.print_stderr(usage_note.replace("_SCRIPT_", __file__))
    return

def main():
    """
    Main routine: parse arguments and perform main processing
    TODO: revise comments
    Note: Used to avoid conflicts with globals (e.g., if this were done at end of script).
    """
    # Initialize
    debug.trace(4, "main(): sys.argv=%s" % sys.argv)

    # Show usage statement if no arguments or if --help specified
    just_tokenize = False
    make_lowercase = False
    just_chunk = False
    show_usage = ((len(sys.argv) == 1) or (sys.argv[1] == "--help"))
    arg_num = 1
    while ((not show_usage) and arg_num < len(sys.argv) and (sys.argv[arg_num][0] == "-")):
        if (sys.argv[arg_num] == "--just-tokenize"):
            just_tokenize = True
        elif (sys.argv[arg_num] == "--lowercase"):
            make_lowercase = True
        elif (sys.argv[arg_num] == "--just-chunk"):
            just_chunk = True
        elif sys.argv[arg_num] == "-":
            break
        else:
            system.print_stderr("Invalid argument: %s" % sys.argv[arg_num])
            show_usage = True
        arg_num += 1
    if (show_usage):
        usage()
        sys.exit()

    # Run the text from each file through the pipeline
    for i in range(arg_num, len(sys.argv)):
        # Input the entire text from the file (or stdin if - specified)
        filename = sys.argv[i]
        input_handle = system.open_file(filename) if (filename != "-") else sys.stdin
        debug.trace_expr(5, filename, input_handle)

        # Analyze the text
        while True:
            text = input_handle.readline() if LINE_MODE else input_handle.read()
            if text == "":
                break
            text = text.strip()
            if just_tokenize:
                # Just tokenize sentence and words
                # Note: text lowercased at end to allow for case clues (e.g., "Dr. Jones")
                for tokenized_line in tokenize_text(text):
                    tokenized_text = " ".join(tokenized_line)
                    if make_lowercase:
                        tokenized_text = tokenized_text.lower()
                    print(tokenized_text)
            elif just_chunk:
                ## OLD: text_proc = FlairTextProc()
                text_proc = create_text_proc(TEXT_PROC)
                if VERBOSE:
                    print(f"text: {text}")
                    print("NP chunks:")
                ## TODO: print(OUTPUT_DELIM.join(text_proc.noun_phrases(text)))
                print(text_proc.noun_phrases(text))
            else:
                # Show complete pipeline step-by-step
                taggings = tokenize_and_tag(text)
                
                # Show the results
                # TODO: show original tags if class-based; put JUST_TAGS support in tag_part_of_speech
                # TODO: flesh out support for verbose mode
                if VERBOSE:
                    print("text: %s" % text)
                if JUST_TAGS:
                    if VERBOSE:
                        print("tokens: %s" % OUTPUT_DELIM.join([w for (w, _POS) in taggings]))
                        print("tags: %s" % OUTPUT_DELIM.join([POS for (_w, POS) in taggings]))
                    else:
                        print(OUTPUT_DELIM.join([POS for (_w, POS) in taggings]))
                else:
                    gh.assertion(VERBOSE)
                    print("taggings: %s" % taggings)
                if SHOW_MISSPELLINGS:
                    misspellings = [w for (w, POS) in taggings if has_spelling_mistake(w)]
                    gh.assertion(VERBOSE)
                    print("misspellings: %s" % misspellings)
                if VERBOSE:
                    print("")

    # Cleanup
    debug.trace(4, "stop %s: %s" % (__file__, debug.timestamp()))
    return

#------------------------------------------------------------------------
# Initialization

init()

if __name__ == "__main__":
    main()



#! /usr/bin/env python3
#
# Test(s) for ../document.py

"""Tests for tfidf document submodule"""

# Standard packages
import math

# Installed modules
import pytest

# Local modules
from mezcla.unittest_wrapper import TestWrapper
from mezcla import debug
from mezcla.tfidf.preprocess import Preprocessor

# Note: Two references are used for the module to be tested:
#    THE_MODULE:                        global module object
import mezcla.tfidf.document as THE_MODULE

# ------------------------------------------------------------------------


class TestPreprocess(TestWrapper):
    """Class for testcase definition"""

    script_module = TestWrapper.get_testing_module_name(__file__, THE_MODULE)
    text = "my man fran is not a man"
    doc = THE_MODULE.Document(text, Preprocessor(gramsize=1, stemmer=lambda x: x))

    @classmethod
    def setUpClass(cls, filename=None, module=None):
        """One-time initialization (i.e., for entire class)"""
        debug.trace(6, f"TestPreprocess.setUpClass(); cls={cls}")
        # note: should do parent processing first
        super().setUpClass(filename, module)
        debug.trace_object(5, cls, label=f"{cls.__class__.__name__} instance")

    @pytest.mark.xfail  # TODO: remove xfail
    def test_01_tf_freq(self):
        """test frequency of ngram in document"""
        # count appareances of different ngrams in document
        man_freq = self.doc.tf_freq("man")
        fran_freq = self.doc.tf_freq("fran")
        peter_freq = self.doc.tf_freq("peter")
        # assert the reported amount is correct
        assert man_freq == 2
        assert fran_freq == 1
        assert peter_freq == 0

    @pytest.mark.xfail  # TODO: remove xfail
    def test_02_tf_raw(self):
        """Tests relative frequency of ngram in document"""
        old_pensalize_singletons = THE_MODULE.PENALIZE_SINGLETONS
        self.monkeypatch.setattr(THE_MODULE, "PENALIZE_SINGLETONS", True)
        assert self.doc.tf_raw("fran") == 0
        assert self.doc.tf_raw("man") == 2 / 7
        self.monkeypatch.setattr(THE_MODULE, "PENALIZE_SINGLETONS", False)
        new_tf_fran = self.doc.tf_raw("fran")
        self.monkeypatch.setattr(
            THE_MODULE, "PENALIZE_SINGLETONS", old_pensalize_singletons
        )
        assert new_tf_fran == 1 / 7

    @pytest.mark.xfail  # TODO: remove xfail
    def test_03_tf_log(self):
        """Tests log frequency of ngram in document"""
        # TODO: point out relation between tf_raw and tf_freq
        old_pensalize_singletons = THE_MODULE.PENALIZE_SINGLETONS
        self.monkeypatch.setattr(THE_MODULE, "PENALIZE_SINGLETONS", True)
        man_log_freq = self.doc.tf_log("man")
        self.monkeypatch.setattr(
            THE_MODULE, "PENALIZE_SINGLETONS", old_pensalize_singletons
        )
        assert man_log_freq == (1 + math.log(2 / 7))

    @pytest.mark.xfail  # TODO: remove xfail
    def test_04_tf_binary(self):
        """Tests binary term frequency"""
        assert self.doc.tf_binary("man") == 1
        assert self.doc.tf_binary("dog") == 0

    @pytest.mark.xfail  # TODO: remove xfail
    def test_04_tf_norm_50(self):
        """Tests double normalized ngram frequency"""
        old_pensalize_singletons = THE_MODULE.PENALIZE_SINGLETONS
        self.monkeypatch.setattr(THE_MODULE, "PENALIZE_SINGLETONS", False)
        man_norm_freq = self.doc.tf_norm_50("man")
        fran_norm_freq = self.doc.tf_norm_50("fran")
        self.monkeypatch.setattr(
            THE_MODULE, "PENALIZE_SINGLETONS", old_pensalize_singletons
        )
        assert man_norm_freq == 0.5 + (0.5 * (2/7) / 2)
        assert fran_norm_freq == 0.5 + (0.5 * (1/7) / 2)

    @pytest.mark.xfail  # TODO: remove xfail
    def test_05_max_raw_frequency(self):
        """Tests doc's max raw frequency"""
        assert self.doc.max_raw_frequency == self.doc.tf_freq("man")

    @pytest.mark.xfail  # TODO: remove xfail
    def test_06_keywordset(self):
        """Tests the set of keywords in the document and their location"""
        keywords = self.doc.keywordset
        text = self.doc.text
        for word, keyword in keywords.items():
            for location_set in keyword.locations:
                location = list(location_set)
                start = location[1]
                end = location[2]
                assert word == text[start:end]
    
    @pytest.mark.xfail # TODO: remove xfail
    def test_07_len(self):
        """Tests the length of the document is calculated correctly"""
        assert len(self.doc) == 7
        
    
    


# ------------------------------------------------------------------------

if __name__ == "__main__":

    debug.trace_current_context()
    ## TODO2: here and elsewhere: invoke_tests(__file__)
    pytest.main([__file__])



#! /usr/bin/env python3
#
# Simple illustration of optional pydantic argument validation.
#
# TODO2: clarify the use of dict's for validation
# TODO3: add support for @validate_call
#

"""Optional pydantic argument validation"""

# Standard modules
import ast
from functools import wraps
import sys

# Installed module
try:
    from pydantic import BaseModel
    import astor
except:
    sys.stderr.write("Warning unable to import astor and/or pydantic\n")
    BaseModel = object
    astor = None

# Local modules
from mezcla import debug
from mezcla import glue_helpers as gh
from mezcla.main import Main
from mezcla import system

# Constants
## OLD: VALIDATE_ARGUMENTS = True
SKIP_VALIDATION = system.getenv_bool(
    "SKIP_VALIDATION", False,
    desc="Skip pydantic-based validation")
VALIDATE_ARGUMENTS = not SKIP_VALIDATION
TL = debug.TL
TMP_PATH = gh.get_temp_dir()
## OLD: TMP_PATH = "/tmp/temp_"

# Arguments for Validate Arguments Script
FILE = "file"
ARG_INPUT_SCRIPT = "input"
ARG_NO_TRANSFORM = "no-transform"
OUTPUT = "output"

def validate_dictionaries(*_decorator_args, **decorator_kwargs):
    """
    Decorator to validate dictionaries with Pydantic models,
    but without changing function definition to keep compatibility
    with mypy, pyright and other.

    Context:

    Basically with Pydantic you can validate dictionary keys and
    values with models, for example:

    class ExpectedDictModel(BaseModel):
        example_key: str

    @validate_call
    fun some_function(some_dict: ExpectedDictModel):
        # Some code here

    But this is not compatible with mypy, pyright and other...

    So, what we do to keep compatibility is to use this decorator
    and pass the parameter to validate with the model like this:

    @validate_dictionaries(some_dict = ExpectedDictModel)
    fun some_function(some_dict: dict):
        # Some code here
    """
    def decorator(func):
        @wraps(func)
        def inner(*func_args, **func_kwargs):
            if not VALIDATE_ARGUMENTS:
                debug.trace(4, "FYI: Ignoring validation for {func}")
                return func(*func_args, **func_kwargs)
            # Validate the dictionary keys and values,
            # specified in the decorator parameters
            for key, decorator_model in decorator_kwargs.items():
                # TODO: assert => debug.assertion
                # Check decorator parameters
                assert issubclass(decorator_model, BaseModel), f"Parameter {key} must be a pydantic.BaseModel class"
                # Check for function parameters
                func_model = func.__annotations__.get(key)
                assert func_model, f"Parameter {key} is missing for validation in function {func.__name__}"
                assert issubclass(func_model, dict), f"Parameter {func_model} must be a dict for validation in function {func.__name__}"
                # Validate the dictionary keys and values
                key_index = list(func.__annotations__).index(key)
                if key_index < len(func_args):
                    func_passed_value = func_args[key_index]
                else:
                    func_passed_value = func_kwargs.get(key)
                decorator_model.model_validate(func_passed_value, strict=True)
            return func(*func_args, **func_kwargs)
        return inner
    return decorator


def add_validate_call_decorator(code):
    """Add the validate_call decorator to all function calls in the code"""
    debug.trace_fmt(6, "add_validate_call_decorator({c})", c=code, max_len=256)

    # Parse the code into AST
    tree = ast.parse(code)

    # Define an import node
    import_node = ast.ImportFrom(
        module='pydantic',
        names=[ast.alias(name='validate_call', asname=None)],
        level=0
    )

    # Define a decorator node
    validate_decorator = ast.Name(id='validate_call', ctx=ast.Load())

    # Add the import statement to the beginning of the AST
    ## TODO2: put after module docstring
    tree.body.insert(0, import_node)

    # List of functions to ignore, avoiding infinite recursion
    to_ignore = [
        'isinstance', 'print', 'validate_call',
    ]

    # Function to recursively traverse the AST
    # pylint: disable=invalid-name
    def visit_Call(node):
        """Visit a Call node"""
        # Standalone functions are represented as Name nodes.
        # Functions from external modules are represented as Attribute nodes.
        if not isinstance(node.func, (ast.Name, ast.Attribute)):
            return node
        # Skip functions to ignore
        if isinstance(node.func, ast.Name) and node.func.id in to_ignore:
            return node
        # Add the decorator to the function call
        node.func = ast.Call(func=validate_decorator, args=[node.func], keywords=[])
        return node

    # Traverse the AST and modify function definitions
    for node in ast.walk(tree):
        if isinstance(node, ast.Call):
            node = visit_Call(node)

    # Convert the modified AST back to Python code
    # TODO2: do this via ast
    modified_code = astor.to_source(tree)
    debug.trace_fmt(5, "add_validate_call_decorator() = {r!r})",
                    r=modified_code, max_len=256)

    return modified_code


class ValidateArgumentsScript(Main):
    """Argument processing class to Validate Arguments"""

    # Class-level member variables for arguments
    # (avoids need for class constructor)
    ## OLD: file = ""
    output = ""

    def setup(self) -> None:
        """Process arguments"""
        ## OLD: self.file = self.get_parsed_argument(FILE, self.file)
        self.output = self.get_parsed_option(OUTPUT, self.output)

    def run_main_step(self) -> None:
        """Process main script"""
        debug.trace(TL.USUAL, f"main(): script={system.real_path(__file__)}")
        ## OLD: code = system.read_file(self.filename)
        filename = (self.filename if (self.filename != "-") else "_stdin_")
        code = self.read_entire_input()
        if not code:
            raise ValueError(f"File {filename} is empty")
        if VALIDATE_ARGUMENTS:
            code = add_validate_call_decorator(code)
        output_filename = self.output if self.output else TMP_PATH + gh.basename(filename)
        system.write_file(
            filename=output_filename,
            text=code
        )
        output = gh.run(f"python3 {output_filename}")
        print(output)


if __name__ == '__main__':
    debug.trace_current_context(level=TL.QUITE_VERBOSE)
    app = ValidateArgumentsScript(
        description=__doc__,
        ## OLD:
        ## positional_arguments = [
        ##     (FILE, 'Python script to run with argument validation')
        ## ],
        text_options=[
            (OUTPUT, 'Output of transformed script'),
        ],
        skip_input=False,
        manual_input=True,
    )
    app.run()



#! /usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Test(s) for ../spell.py
# ^ TODO: More?
#
# Notes:
# - For debugging the tested script, the ALLOW_SUBCOMMAND_TRACING environment
#   option shows tracing output normally suppressed by  unittest_wrapper.py.
# - This can be run as follows:
#   $ PYTHONPATH=".:$PYTHONPATH" python ./mezcla/tests/test_bing_search.py
# - Spellcheck requires nltk and hunspell for multi-language support
#   Language specific hunspell can be installed as hunspell-LANG_CODE (e.g. hunspell-ne -> Nepali)
#   $ apt-get install hunspell-ne 
#
# TODO1:
# - Rework tests following test_spell_EN
# - Carefully review tests/template.py.
# TODO3:
# - Create test helpers to cut down on all the redundant code.
#   ex: def check_spelling(self, lang, text, bad):
#       """Run spelling over TEXT in TEXT looking for BAD words""
#       output = self.run_script(env_options=f"SPELL_LANG={lang}", data_file=self.create_temp_file(text))
#       assert(system.intersection(text.split(), bad.split()))
# TODO4:
# - Reword "This test can take some time or may have missing libraries" so that just
#   covers RUN_SLOW_TESTS; Check separately for missing libraries.
#

"""Tests for spell module"""
# Standard packages
## OLD: import re

# Installed packages
import pytest
# import unittest

# Local packages
from mezcla import debug
from mezcla import glue_helpers as gh
from mezcla import system
from mezcla.unittest_wrapper import TestWrapper, invoke_tests
## TODO: from mezcla.unittest_wrapper import trap_exception

# Note: Two references are used for the module to be tested:
#    THE_MODULE:	    global module object
#    TestIt.script_module:              path to file
try:
    # note: requires enchant-2 library package under Ubuntu (besides pyenchant)
    import mezcla.spell as THE_MODULE
except:
    THE_MODULE = None
    system.print_exception_info("spell import")

# Since importing doesn't work properly, a temporary fix is specifying the path
# import mezcla.spell as THE_MODULE
# *** It works fine: please follow the example in test/template.py better! ***
## BAD
## PATH1 = f'$PWD/mezcla/spell.py'
## PATH2 = f'$PWD/spell.py'
## PATH3 = f'../spell.py'
## PWD_COMMAND = f'echo $PWD'
## echo_PWD = gh.run(PWD_COMMAND)
## if (echo_PWD.endswith('/mezcla/mezcla')):
##     SPELL_PATH = PATH2
## elif (echo_PWD.endswith('/mezcla')):
##     SPELL_PATH = PATH1
## else:
##     SPELL_PATH = PATH3
##
## NOTE: the following is a temporary workaround for the bad tests
## TODO1: use run_script as in the template (and in most other tests elsewhere)!
SPELL_PATH = gh.resolve_path("../spell.py")

## Environment Variables
RUN_SLOW_TESTS = system.getenv_bool(
    "RUN_SLOW_TESTS", 
    False,
    description="Run tests that can a while to run"
    )

# @pytest.mark.skipif(not THE_MODULE, reason="Problem loading spell.py: check requirements")

class SpellFiles(TestWrapper):
    """Class for testcase definition"""
    script_module = TestWrapper.get_testing_module_name(__file__, THE_MODULE)
    use_temp_base_dir = True    # treat TEMP_BASE as directory
         
    # NOTE 1: For test_phrase, song lyrics are used
    # NOTE 2: The content in test_phase error MUST be all lowercase
    # NOTE 3: Standard assertion case: assert (output in test_phrase_error and len(output) != 0)
    # TODO 1: Find a method to not create any external filess when test_spell_query_LL functions used
    # TODO 2: A function for test_run_command replacement (optional)

    # Function A: Helper Function reduces the amount of code to be written
    # @pytest.mark.skip
    def helper_spell(
        self, 
        lang_code:str="en_EN", 
        phrase:str="Hello World", 
        batch_file_path:str="-",
        ):
        """Helper function for test_spell.py"""
        debug.trace(4, f"\nhelper_spell(); self={self}")
        
        data_file = self.create_temp_file(contents=phrase) if batch_file_path == "-" else batch_file_path
        output = self.run_script(
            env_options=f"SPELL_LANG={lang_code}",
            data_file=data_file,
        )

        debug.trace_expr(5, output, lang_code, phrase, batch_file_path)      
        return output

        ## OLD    
        # if batch_file_path == "-" and query_like == False:
        #     command = f'echo "{phrase}" | SPELL_LANG={lang_code} {SPELL_PATH} {batch_file_path}'
        # elif query_like:
        #     temp_file = self.create_temp_file(contents=phrase) if query_like else batch_file_path
        #     command = f'SPELL_LANG={lang_code} {SPELL_PATH} {temp_file}'
        # else:
        #     command = f'SPELL_LANG={lang_code} {SPELL_PATH} {batch_file_path}'
        
        # output = gh.run(command)
    
    ## OLD: Merged to helper_spell()
    # # Function B: Helper Function for query-like tests (skipped by default)
    # # @pytest.mark.skip
    # def helper_spell_tempfile(
    #     self,
    #     lang_code:str="en_EN",
    #     phrase:str="Hello World"
    #     ):
    #     """Helper function using tempfile"""
    #     debug.trace(4, f"\nhelper_spell_tempfile(); self={self}")

    #     ## OLD: 
    #     # temp_phrase = f"{self.temp_file}.phrase"
    #     # test_run_command_1 = f'echo "{phrase}" > {temp_phrase}'
    #     # test_run_command_2 = f'SPELL_LANG={lang_code} {SPELL_PATH} {temp_phrase} > {self.temp_file}'
        
    #     # gh.issue(test_run_command_1)
    #     # output = gh.run(test_run_command_2)
    #     # output = self.run_script(self.temp_file)
        
    #     temp_file = self.create_temp_file(contents=phrase)
    #     command = f'SPELL_LANG={lang_code} {SPELL_PATH} {temp_file}'
    #     output = gh.run(command)
    #     debug.trace_expr(5, output, lang_code, phrase)
    #     return output
    
    # @pytest.mark.skip
    @pytest.mark.xfail                   # TODO: remove xfail
    def test_spell_default(self):
        """Ensure test_spell_default [English] works as expected"""
        debug.trace(4, f"\ntest_spell_default(); self={self}")
        
        test_phrase = "One kiss is all it tajkes"
        test_phrase_error = "tajkes"
        output = self.helper_spell(phrase=test_phrase)

        debug.trace_expr(5, output, test_phrase_error)
        assert (output == test_phrase_error and len(output) != 0)

    @pytest.mark.xfail                   # TODO: remove xfail
    # @trap_exception                      # TODO: remove when debugged
    # @pytest.mark.xfail                   # TODO: remove xfail
    def test_spell_EN(self):
        """Ensure test_spell_EN [English] works as expected"""
        debug.trace(4, f"\ntest_spell_EN(); self={self}")
        
        test_lang = "en_EN"
        test_phrase = "You givc me that feeling"
        test_phrase_error = "givc"

        ## TODO: maldito vs code so awkward to use!
        output = self.helper_spell(test_lang, test_phrase)
        debug.trace_expr(5, output, test_phrase_error)
        assert (output in test_phrase_error and len(output) != 0)
        return

    @pytest.mark.skipif(not RUN_SLOW_TESTS, reason="This test can take some time or may have missing libraries")               
    @pytest.mark.xfail                   # TODO: remove xfail
    def test_spell_ES(self):
        """Ensure test_spell_ES [Spanish] works as expected"""
        debug.trace(4, f"\ntest_spell_ES(); self={self}")
        
        test_lang = "es_ES"
        test_phrase = "Yo te miro y se me corta la respiraciónqq"
        test_phrase_error = "respiraciónqq"
        output = self.helper_spell(test_lang, test_phrase)
        
        debug.trace_expr(5, output, test_phrase_error)
        assert (output in test_phrase and len(output) != 0)
        return

    @pytest.mark.skipif(not RUN_SLOW_TESTS, reason="This test can take some time or may have missing libraries")               
    @pytest.mark.xfail                   # TODO: remove xfail
    def test_spell_NE(self):
        """Ensure test_spell_NE y[Nepali] works as expected"""
        debug.trace(4, f"\ntest_spell_NE(); self={self}")
        
        test_lang = "ne_NE"
        test_phrase = "थाहा छैन तिमीर को हो मेरो"
        test_phrase_error = "तिमीर"
        output = self.helper_spell(lang_code=test_lang, phrase=test_phrase)
        
        debug.trace_expr(5, output, test_phrase_error)
        assert (output in test_phrase and len(output) != 0)
        return

    @pytest.mark.skipif(not RUN_SLOW_TESTS, reason="This test can take some time or may have missing libraries")
    @pytest.mark.xfail                   # TODO: remove xfail
    def test_spell_AR(self):
        """Ensure test_spell_AR [Arabic] works as expected"""
        debug.trace(4, f"test_spell_AR(); self={self}")

        test_lang = "ar_AR"
        test_phrase = "سعلينا الهوى وغنّا"
        output = self.helper_spell(test_lang, test_phrase)
        
        debug.trace_expr(5, output, test_phrase)
        assert (output in test_phrase and len(output)!=0)
        return
    
    @pytest.mark.skipif(not RUN_SLOW_TESTS, reason="This test can take some time or may have missing libraries")
    @pytest.mark.xfail                   # TODO: remove xfail
    def test_spell_RU(self):
        """Ensure test_spell_RU [Russian] works as expected"""
        debug.trace(4, f"test_spell_RU(); self={self}")
        
        test_lang = "ru_RU"
        test_phrase = "Поплыли туманыны над рекой"
        output = self.helper_spell(test_lang, test_phrase)
        
        debug.trace_expr(5, output, test_phrase)
        assert (output in test_phrase and len(output) != 0)
        return
    
    @pytest.mark.skipif(not RUN_SLOW_TESTS, reason="This test can take some time or may have missing libraries")
    def test_spell_query_EN(self):
        """Ensure test_spell_query_EN works as expected"""
        debug.trace(4, f"test_spell_query_EN(); self={self}")

        test_lang = "en_EN"
        test_phrase = "Because I am lostr in the way you move"
        test_phrase_error = "lostr"
        output = self.helper_spell(test_lang, test_phrase)

        debug.trace_expr(5, output, test_phrase_error)
        assert (output == test_phrase_error and len(output) != 0)
        return

    @pytest.mark.skipif(not RUN_SLOW_TESTS, reason="This test can take some time or may have missing libraries")
    def test_spell_query_ES(self):
        """Ensure test_spell_query_ES works as expected"""
        debug.trace(4, f"test_spell_query_ES(); self={self}")

        
        test_lang = "es_ES"
        test_phrase = "Me dijeron que te estás casandoi"
        test_phrase_error = "casandoi"
        output = self.helper_spell(test_lang, test_phrase)
        
        debug.trace_expr(5, output, test_phrase_error)
        assert (output == test_phrase_error and len(output) != 0)
        return

    @pytest.mark.skipif(not RUN_SLOW_TESTS, reason="This test can take some time or may have missing libraries")
    def test_spell_query_NE(self):
        """Ensure test_spell_query_NE works as expected"""
        debug.trace(4, f"test_spell_query_NE(); self={self}")

        test_lang = "ne_NE"
        test_phrase = "तिमी नै अबप मेरो झुल्केको बिहानी"
        test_phrase_error = "अबप"
        output = self.helper_spell(test_lang, test_phrase)

        debug.trace_expr(5, output, test_phrase_error)
        assert (output in test_phrase and len(output) != 0)
        return

    @pytest.mark.skipif(not RUN_SLOW_TESTS, reason="This test can take some time or may have missing libraries")
    @pytest.mark.xfail                   # TODO: remove xfail
    def test_spell_default_batch(self):
        """Ensure test_spell_default_batch [English] works as expected"""
        debug.trace(4, f"test_spell_default_branch(); self={self}")
        
        testfile_path = gh.resolve_path("./resources/spell-py-en.list")
        output = (self.helper_spell(batch_file_path=testfile_path)).split("\n")
        
        debug.trace_expr(5, output)
        assert (output != [] and len(output) > 5)
        return
    
    @pytest.mark.skipif(not RUN_SLOW_TESTS, reason="This test can take some time or may have missing libraries")
    @pytest.mark.xfail                   # TODO: remove xfail
    def test_spell_ES_batch(self):
        """Ensure test_spell_ES_batch [Spanish] works as expected"""
        debug.trace(4, f"test_spell_ES_batch(); self={self}")
        
        test_lang = "es_ES"
        testfile_path = gh.resolve_path("./resources/spell-py-es.list")
        output = (self.helper_spell(lang_code=test_lang, batch_file_path=testfile_path)).split("\n")
        
        debug.trace_expr(5, output)
        assert (output != [] and len(output) > 3)
        return
    
    @pytest.mark.skipif(not RUN_SLOW_TESTS, reason="This test can take some time or may have missing libraries")
    @pytest.mark.xfail                   # TODO: remove xfail
    def test_spell_RU_batch(self):
        """Ensure test_spell_RU_batch [Russian] works as expected"""
        debug.trace(4, f"test_spell_RU_batch(); self={self}")
        
        test_lang = "ru_RU"
        testfile_path = gh.resolve_path("./resources/spell-py-ru.list")
        output = (self.helper_spell(lang_code=test_lang, batch_file_path=testfile_path)).split("\n")
        
        debug.trace_expr(5, output)
        assert (output != [] and len(output) > 3)
        return
    
    @pytest.mark.skipif(not RUN_SLOW_TESTS, reason="This test can take some time or may have missing libraries")
    @pytest.mark.xfail                   # TODO: remove xfail
    def test_spell_AR_batch(self):
        """Ensure test_spell_AR_batch [Arabic] works as expected"""
        debug.trace(4, f"test_spell_AR_batch(); self={self}")
        
        test_lang = "ar_AR"
        testfile_path = gh.resolve_path("./resources/spell-py-ar.list")
        output = (self.helper_spell(lang_code=test_lang, batch_file_path=testfile_path)).split("\n")
        
        debug.trace_expr(5, output)
        assert (output != "" and len(output)>5)
        return

    @pytest.mark.xfail                   # TODO: remove xfail
    def test_spell_LANG_suggest(self):
        """Ensure test_spell_LANG_suggest works as expected"""
        debug.trace(4, f"test_suggest_LANG(); self={self}")
        self.do_assert(False, "TODO: code test")

if __name__ == '__main__':
    debug.trace_current_context()
    invoke_tests(__file__)



#! /usr/bin/env python3
# -*- coding: utf-8 -*
#
# Test(s) for ../hugging_face_translation.py
#

"""Tests for hugging_face_translation module"""

# Standard modules
## TODO: from collections import defaultdict

# Installed modules
import pytest

# Local modules
from mezcla.unittest_wrapper import TestWrapper
from mezcla import debug
from mezcla import glue_helpers as gh
from mezcla.my_regex import my_re
from mezcla import system
from mezcla.tests.common_module import SKIP_SLOW_TESTS, SKIP_SLOW_REASON

# Note: Two references are used for the module to be tested:
#    THE_MODULE:                        global module object
#    TestIt.script_module:              path to file
import mezcla.examples.hugging_face_translation as THE_MODULE
#
# Note: sanity test for customization (TODO: remove if desired)
if not my_re.search(__file__, r"\btemplate.py$"):
    debug.assertion("mezcla.template" not in str(THE_MODULE))

## TODO3: fix temporary hack to use gh.form_path here and gh.resolve_path
# TEMP FIX: Path specification for mezcla scripts
SCRIPT = "hugging_face_translation.py"
PATH = SCRIPT
PATH1 = f"$PWD/mezcla/examples/{SCRIPT}"
PATH2 = f"$PWD/examples/{SCRIPT}"
PATH3 = f"$PWD/{SCRIPT}"
PATH4 = f"../{SCRIPT}"
PWD_COMMAND = "echo $PWD"
echo_pwd = gh.run(PWD_COMMAND)
if echo_pwd.endswith("/mezcla/mezcla/examples"):
    HF_TRANSLATION_PATH = PATH3
elif echo_pwd.endswith("/mezcla/mezcla"):
    HF_TRANSLATION_PATH = PATH2
elif echo_pwd.endswith("/mezcla"):
    HF_TRANSLATION_PATH = PATH1
else:
    HF_TRANSLATION_PATH = PATH
## TODO3: rework HF_TRANSLATION_PATH usage below using run_script
if not system.file_exists(HF_TRANSLATION_PATH):
    HF_TRANSLATION_PATH = gh.resolve_path(SCRIPT)
debug.assertion(system.file_exists(HF_TRANSLATION_PATH))
                
#------------------------------------------------------------------------

class TestIt(TestWrapper):
    """Class for command-line based testcase definition"""
    script_module = TestWrapper.get_testing_module_name(__file__, THE_MODULE)

    @pytest.mark.xfail                   # TODO: remove xfail
    def test_data_file(self):
        """Tests run_script w/ data file"""
        debug.trace(4, f"TestIt.test_data_file(); self={self}")
        
        ## OLD:
        ## system.write_file(self.temp_file, "Hi, Paul")
        ## system.write_file(self.temp_file, "Hi, Paul")
        ## output = self.run_script(options="", env_options="FROM=es TO=en",
        ##                          data_file=self.temp_file)
        ## self.do_assert("hola" in output.lower())
        ## self.do_assert("pablo" in output.lower())
        ##
        ## NOTE: Translation is fickle with proper names so use proper noun phrase.
        system.write_file(self.temp_file, "John the Baptist")
        output = self.run_script(options="", env_options="FROM=en TO=es",
                                 data_file=self.temp_file)
        self.do_assert("juan" in output.lower())
        self.do_assert("bautista" in output.lower())
        return
    
    ## NEW: Revised tests from mezcla/tests/test_huggingface_translation.py
    
    ## Test 1 - Default Run: ES -> EN
    @pytest.mark.skipif(SKIP_SLOW_TESTS, reason=SKIP_SLOW_REASON)
    def test_translation_default(self):
        """Ensures that test_translation_default works properly"""
        debug.trace(4, "test_translation_default()")
        source_sentence = "Hola Soy Dora."
        target_sentence = "Hi, I'm Dora."
        command = f"echo {source_sentence} | python3 {HF_TRANSLATION_PATH} - 2> /dev/null"
        command_output = gh.run(command)
        assert(command_output == target_sentence)
        return
    
    ## Test 2 - Translation I: NOT_EN -> EN (e.g. Japanese)
    @pytest.mark.skipif(SKIP_SLOW_TESTS, reason=SKIP_SLOW_REASON)
    def test_translation_ja2en(self):
        """Ensures that test_translation_ja2en works properly"""
        debug.trace(4, "test_translation_ja2en()")
        source_lang, target_lang = "ja", "en"
        source_sentence = "かわいいですね。"
        target_sentence = "It's cute."
        command = f"echo {source_sentence} | FROM={source_lang} TO={target_lang} python3 {HF_TRANSLATION_PATH} - 2> /dev/null"
        command_output = gh.run(command)
        assert(target_sentence in command_output)
        return
    
    ## Test 3 - Translation II: EN -> NON_EN (e.g. French)
    @pytest.mark.skipif(SKIP_SLOW_TESTS, reason=SKIP_SLOW_REASON)
    def test_translation_en2fr(self):
        """Ensures that test_translation_en2fr works properly"""
        debug.trace(4, "test_translation_en2fr()")
        source_lang, target_lang = "en", "fr"
        ## OLD: Use of string with quotes halts the test
        # source_sentence = "It's cute."
        # target_sentence = "C'est mignon."
        source_sentence = "The people love croissant."
        target_sentence = "Les gens aiment les croissants."
        command = f"echo {source_sentence} | FROM={source_lang} TO={target_lang} python3 {HF_TRANSLATION_PATH} - 2> /dev/null"
        command_output = gh.run(command)
        assert(target_sentence == command_output)
        return
    
    ## Test 4 - Translation III: NON_EN -> NON_EN (e.g. Russian to Arabic)
    @pytest.mark.skipif(SKIP_SLOW_TESTS, reason=SKIP_SLOW_REASON)
    def test_translation_ru2ar(self):
        """Ensures that test_translation_ru2ar works properly"""
        debug.trace(4, "test_translation_ru2ar()")
        source_lang, target_lang = "ru", "ar"
        # Literal Translation: I love potato juice.
        source_sentence = "Я люблю картофельный сок."
        target_sentence = "أنا أحب عصير البطاطس."
        command = f"echo {source_sentence} | FROM={source_lang} TO={target_lang} python3 {HF_TRANSLATION_PATH} - 2> /dev/null"
        command_output = gh.run(command)
        assert(target_sentence == command_output)
        return

    ## Test 5 - Translation IV: Using a different model (e.g. t5-small)
    ## NOTE: Default Translation: EN (English) -> DE (German)
    @pytest.mark.skipif(SKIP_SLOW_TESTS, reason=SKIP_SLOW_REASON)
    def test_translation_t5small(self):
        """Ensures that test_translation_t5small works properly"""
        debug.trace(4, "test_translation_t5small()")
        mt_model = "t5-small"
        source_sentence = "Dortmund is black and yellow."
        target_sentence = "Dortmund ist Schwarz und yellow."
        command = f"echo {source_sentence} | python3 {HF_TRANSLATION_PATH} --model {mt_model} - 2> /dev/null"
        command_output = gh.run(command)
        assert(command_output in target_sentence)
        return

#------------------------------------------------------------------------

if __name__ == '__main__':
    debug.trace_current_context()
    pytest.main([__file__])



#! /usr/bin/env python3
#
# Test(s) for ../keras_param_search.py
#
# Notes:
# - For debugging the tested script, the ALLOW_SUBCOMMAND_TRACING environment
#   option shows tracing output normally suppressed by  unittest_wrapper.py.
# - This can be run as follows:
#   $ PYTHONPATH=".:$PYTHONPATH" python ./mezcla/tests/test_keras_param_search.py
#

"""Tests for keras_param_search module"""

# Standard packages
## TODO: import re

# Installed packages
import pytest

# Local packages
from mezcla import debug
from mezcla import system
from mezcla import text_utils
from mezcla.unittest_wrapper import TestWrapper, invoke_tests

# Load module conditionally because not part of default installation.
# Note: Two references are used for the module to be tested:
#    THE_MODULE:               module object (e.g., <module 'mezcla.main' ...>)
#    TestIt.script_module:     dotted module path (e.g., "mezcla.main")
try:
    import mezcla.keras_param_search as THE_MODULE
except:
    THE_MODULE = None
    debug.trace_exception(1, "keras_param_search import")

class TestKerasParamSearch(TestWrapper):
    """Class for testcase definition"""
    # note: script_module used in argument parsing sanity check (e.g., --help)
    script_module = TestWrapper.get_testing_module_name(__file__, THE_MODULE)

    @pytest.mark.skipif(not THE_MODULE, reason="Unable to load module")
    def test_round3(self):
        """Ensure round3 works as expected"""
        debug.trace(4, f"test_round3(); self={self}")
        assert THE_MODULE.round3(1.0001) == 1.000
        assert THE_MODULE.round3(1.001) == 1.001

    @pytest.mark.skipif(not THE_MODULE, reason="Unable to load module")
    def test_non_negative(self):
        """Ensure non_negative works as expected"""
        debug.trace(4, f"test_non_negative(); self={self}")
        test_num = 32.86209423
        assert THE_MODULE.non_negative(test_num)

    @pytest.mark.skipif(not THE_MODULE, reason="Unable to load module")
    def test_round3_too(self):
        """Ensure test_round3 works as expected"""
        debug.trace(4, "test_non_negative()")
        test_num = 3.1415926543
        assert THE_MODULE.round3(test_num) == 3.142

    @pytest.mark.skipif(not THE_MODULE, reason="Unable to load module")
    def test_create_feature_mapping(self):
        """Ensure create_feature_mapping works as expected"""
        debug.trace(4, f"test_create_feature_mapping(); self={self}")
        assert THE_MODULE.create_feature_mapping(['c', 'b', 'b', 'a']) == {'c':0, 'b':1, 'a':2}

    @pytest.mark.skipif(not THE_MODULE, reason="Unable to load module")
    @pytest.mark.xfail
    def test_create_keras_model(self):
        """Ensure create_keras_model works as expected"""
        debug.trace(4, "test_create_keras_model()")
        model = THE_MODULE.create_keras_model(
            num_input_features = 100, 
            num_classes = 2,
            hidden_units = text_utils.getenv_ints("HIDDEN_UNITS", "20 30"))
        # Make sure model size makes sense
        # example: [100, 20, 20, 30, 30, 1] for following
        model_param_lens = [len(w) for w in model.get_weights()]
        debug.trace_expr(5, model_param_lens)
        assert system.intersection(model_param_lens, [100, 20, 30])

    @pytest.mark.xfail
    def test_MyKerasClassifier_check_params(self):
        """Ensure MyKerasClassifier.check_params works as expected"""
        debug.trace(4, "test_MyKerasClassifier_check_params()")
        ## TODO: test MyKerasClassifier class
        assert False, "TODO: code test"
        
    @pytest.mark.xfail
    def test_main(self):
        """Check main routine"""
        debug.trace(4, "test_main()")
        ## TODO: test main
        assert False, "TODO: code test"

#------------------------------------------------------------------------

if __name__ == '__main__':
    debug.trace_current_context()
    invoke_tests(__file__)



#! /usr/bin/env python3
#
# Runs regression over specifid data file using R.
#
# Notes:
# - Automaes following process
#     data = read.delim("sample-deduped-jobs-title.score.data")
#     fit = lm(RELDIFF ~ REL + NGRAM + POS + CAT + EDIT + FREQ, data=data)
# - R invocation options:
#     --vanilla: Combine --no-save, --no-restore, --no-site-file,  
#                --no-init-file  and --no-environ
#     --verbose:
#
# TODO2:
# - Add regression via pandas and/or sklearn as a sanity check.
#
#-------------------------------------------------------------------------------
# Sample regression output:
#
# R version 2.15.1 (2012-06-22) -- "Roasted Marshmallows"
# ...
# Call:
# lm(formula = RELDIFF ~ REL_SC + NGRAMSC + POS_SC + CAT_SC + EDIT_SC +
#     FREQ_SC, data = data)
#
# Residuals:
#      Min       1Q   Median       3Q      Max
# -0.66196 -0.19160 -0.00842  0.18282  0.79553
#
# Coefficients:
#              Estimate Std. Error t value Pr(>|t|)
# (Intercept) -0.023724   0.008609  -2.756  0.00588 **
# REL_SC       4.052147   0.112756  35.937  < 2e-16 ***
# NGRAMSC     -0.222896   0.025619  -8.700  < 2e-16 ***
# POS_SC       0.106062   0.066014   1.607  0.10818
# CAT_SC       0.161975   0.019235   8.421  < 2e-16 ***
# EDIT_SC      0.274574   0.009212  29.807  < 2e-16 ***
# FREQ_SC     -0.429607   0.019751 -21.751  < 2e-16 ***
# ---
# Signif. codes:  0 '***' 0.001 '**' 0.01 '*' 0.05 '.' 0.1 ' ' 1
#
# Residual standard error: 0.2462 on 5898 degrees of freedom
# Multiple R-squared: 0.4068,     Adjusted R-squared: 0.4062
# F-statistic: 674.1 on 6 and 5898 DF,  p-value: < 2.2e-16
#-------------------------------------------------------------------------------
# TODO:
# - Add option to include diagnostics plot:
#     save_par = par(no.readonly = TRUE)
#     par(mfrow = c(2, 2))
#     plot(fit)
#     par(save_par)
# - Allow for symbolic column specification.
# - Add option to use SciPy-based regression.
#

"""Performs linear regression"""

# Standard modules
import argparse
import re
import sys
import tempfile

# Installed modules
from mezcla import debug
from mezcla import tpo_common as tpo
from mezcla import glue_helpers as gh
from mezcla import system
from mezcla.text_utils import extract_string_list

# Constants
TL = debug.TL

# Environment options
TAB_SEPARATED_FEATURES = tpo.getenv_boolean("TAB_SEPARATED_FEATURES", False,
                                            "Separate feature values by tabs")
## OLD: DEFAULT_FEATURE_DELIM = ", " if not TAB_SEPARATED_FEATURES else "\t"
## TODO3: DEFAULT_FEATURE_DELIM = ", *"
DEFAULT_FEATURE_DELIM = "," if not TAB_SEPARATED_FEATURES else "\t"
FEATURE_DELIM = tpo.getenv_text("FEATURE_DELIM", DEFAULT_FEATURE_DELIM, 
                                "Delimiter for output features")
debug.assertion(len(FEATURE_DELIM) == 1)

TEMP_BASE = tpo.getenv_text("TEMP_BASE", tempfile.NamedTemporaryFile().name,
                            "Basename for temporary file(s)")

def main():
    """Entry point for script"""
    tpo.debug_print("main(): sys.argv=%s" % sys.argv, 4)

    # Check command-line arguments
    env_options = tpo.formatted_environment_option_descriptions(indent="  ")
    usage_description = tpo.format("""
Runs regression via R using lm function.
Notes:
- The column numbers are 1-based.
- Install R via apt-get (e.g., "apt-get install --yes --fix-missing r-base").
- The following environment options are available:
  {env}
""", env=env_options)
    parser = argparse.ArgumentParser(description=usage_description, \
                        formatter_class=argparse.RawDescriptionHelpFormatter)
    parser.add_argument("--independent-cols", default="",
                        help="Column numbers for independent variables (Xi)")
    parser.add_argument("--dependent-col", type=int, default=1,
                        help="Column number for dependent variable (Y); -1 for last",)
    parser.add_argument("--logit", default=False, action='store_true',
                        help="Perform logistic regression")
    parser.add_argument("--quasi", default=False, action='store_true',
                        help="Use quasi- version of family for glm (e.g., quasibinomial)")
    parser.add_argument("--plot", default=False, action='store_true',
                        help="Output plot (to basename + .png)")
    parser.add_argument("--verbose", default=False, action='store_true',
                        help="Verbose output mode")
    parser.add_argument("--anova", default=False, action='store_true',
                        help="Run anova instead")
    parser.add_argument("--labels", default="",
                        help="Labels for columns--comma separated")
    parser.add_argument("filename", help="Input filename")
    args = vars(parser.parse_args())
    tpo.debug_print("args = %s" % args, 5)
    dependent_col_num = args['dependent_col']
    col_list = args['independent_cols'].replace(",", " ").split()
    independent_col_nums = [tpo.safe_int(c) for c in col_list]
    filename = args['filename']
    produce_plot = args['plot']
    verbose = args['verbose']
    use_logit = args['logit']
    use_quasi = args['quasi']
    use_anova = args['anova']
    labels = args['labels']
    tpo.trace_array(independent_col_nums, 5, "ind")

    # Output stdin to file
    # TODO2: use main to make this automatic
    if filename == "-":
        filename = f"{TEMP_BASE}.in"
        data = f"{labels}\n" if labels else ""
        data += system.read_all_stdin()
        system.write_file(filename, data)
    
    # Extract column names from data file (first line)
    # note: dashses should be converted to underscore for sake of R
    #
    column_names = []
    if labels:
        column_names = extract_string_list(labels)
    else:
        header_line = gh.run("head -1 {f}", f=filename)
        gh.assertion("-" not in header_line)
        ## TODO: header_line = gh.run("head -1 {f}", f=filename).replace("-", "_")
        column_names = header_line.split(FEATURE_DELIM)
    tpo.trace_array(column_names, 5, "column_names")
    #
    if dependent_col_num == -1:
        dependent_col_num = len(column_names)
    gh.assertion(0 < dependent_col_num <= len(column_names))
    #
    # Work out defaults for independent columns
    if not independent_col_nums:
        independent_col_nums = [c for c in range(1, 1 + len(column_names))
                                if c != dependent_col_num]
        tpo.debug_format("independent_col_nums = {nums}", 3,
                         nums=independent_col_nums)

    # Resolve column labels
    def get_column_name(num):
        """Returns symbolic name for column"""
        gh.assertion(num <= len(column_names))
        tpo.trace_array(column_names, 5, "(local) column_names")
        label = column_names[num - 1] if num <= len(column_names) else "???"
        tpo.debug_format("get_column_name({n}) => {l}", 5, n=num, l=label)
        return label
    #
    dependent_label = get_column_name(dependent_col_num)
    independent_labels = [get_column_name(c) for c in independent_col_nums]

    # Create R script for regression
    lm_function = "lm"
    analysis_fn = "summary"
    lm_type = "linear"
    lm_qual = ""
    lm_misc = ""
    if use_logit:
        lm_function = "glm" 
        lm_type = "logistic"
        lm_qual = "generalized "
        family = "binomial"
        if use_quasi:
            family = "quasi" + family
        lm_misc = ", family={f}".format(f=family)
    if use_anova:
        analysis_fn = "anova"
    r_template = """
        # Run {type} regresssion (i.e., {qual}linear model)
        data = read.delim("{f}", sep="{delim}")
        fit = {func}({dep} ~ {ind_spec}, data=data{misc})
        {analysis_fn}(fit)
        """.format(func=lm_function, type=lm_type, qual=lm_qual, misc=lm_misc,
                   f=filename, dep=dependent_label, delim=FEATURE_DELIM,
                   analysis_fn=analysis_fn, ind_spec=(" + ".join(independent_labels)))

    # Optionally produce diagnostic plot
    # TODO: parameterize image type, dimensions, etc.
    if produce_plot:
        image_file = gh.remove_extension(filename, ".data") + ".png"
        r_template += """

            # Produce diagnostics plot
            png(filename="{img_file}", width=1920, height=1080)
            par(mfrow = c(2, 2))
            plot(fit)
            dev.off()
            """.format(img_file=image_file)
        print("Outputting plot to " + image_file)

    # Output script to temp file
    r_file = TEMP_BASE + ".r"
    gh.write_file(r_file, r_template)

    # Run R script in basic mode (i.e., "vanilla" environment)
    options = "--quiet"
    if verbose:
        options += " --verbose"
    output = gh.run("R --vanilla {opt} < {f}", f=r_file, opt=options)
    if use_anova:
        gh.assertion(re.search("Analysis of Variance Table.*F.value", output,
                               re.MULTILINE|re.DOTALL))
    else:
        gh.assertion(re.search("Residuals.*Coefficients.*Intercept", output, \
                               re.MULTILINE|re.DOTALL))
    output = re.sub("^.*Call:", "", output, re.MULTILINE|re.DOTALL)
    print(output)

    return

#------------------------------------------------------------------------

if __name__ == '__main__':
    debug.trace_current_context(level=TL.QUITE_VERBOSE)
    main()
